"""
TODO Later Should Klein: Controleer of de buitenteenlijn ook echt aan de buitenzijde is.

"""

from __future__ import annotations
from InquirerPy import inquirer
from typing import TYPE_CHECKING, Optional
import os
import warnings
from shapely import LineString, MultiLineString
from geopandas import GeoDataFrame, read_file
import fiona
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def added_buitenteenlijn(app_settings: ApplicationSettings) -> bool:
    layers = fiona.listlayers(app_settings.geopackage_filepath)

    if "buitenteenlijn" in layers:
        print(BColors.OKBLUE, f"✔  Buitenteenlijn al toegevoegd.", BColors.ENDC)
        return True

    request_buitenteenlijn_filepath(app_settings=app_settings)
    return True


def request_buitenteenlijn_filepath(app_settings: ApplicationSettings):

    # Request filepath
    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar de geopackage/shapefile/geodatabase waarin de "
                    "buitenteen lijnen zitten.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not (filepath.endswith(".gpkg") or filepath.endswith(".shp") or filepath.endswith(".gdb")):
            print(BColors.WARNING, f"Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                                   f"eindigt op de extensie .{filepath.split(sep='.')[-1]}.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    # Import data
    if filepath.endswith(".shp"):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
            gdf: GeoDataFrame = read_file(filepath)
    elif filepath.endswith(".gpkg"):
        gdf: GeoDataFrame = import_from_geopackage(filepath=filepath)
    elif filepath.endswith(".gdb"):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
            gdf: GeoDataFrame = import_from_geodatabase(filepath=filepath)
    else:
        raise NotImplementedError(
            f"Applicatie vroegtijdig afgesloten: Een {filepath.split(sep='.')[-1]}-bestand is niet geïmplementeerd.")

    # Confirm all are points
    all_geometries_are_points = gdf.geometry.apply(
        lambda geom: isinstance(geom, LineString) or isinstance(geom, MultiLineString)).all()
    if not all_geometries_are_points:
        print(BColors.WARNING, f"Het geïmporteerde bestand bestaat niet (volledig) uit lijnen, maar ook uit "
                               f"andere typen geometrie. Enkel lijnen zijn toegestaan.", BColors.ENDC)
        request_buitenteenlijn_filepath(app_settings=app_settings)

    # Add buitenteenlijn
    gdf_to_add = gdf[["geometry"]]
    gdf_to_add.to_file(app_settings.geopackage_filepath, layer="buitenteenlijn", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Buitenteenlijn toegevoegd.", BColors.ENDC)


def import_from_geodatabase(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de layer waarin de buitenteenlijn staat. Type 'listlayers' om "
                    "een overzicht te krijgen van de geodatabase-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geodatabase: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geodatabase: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf


def import_from_geopackage(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de laag met de buitenteen lijnen. Type 'listlayers' om "
                    "een overzicht te krijgen van de geopackage-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geopackage: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geopackage: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf



from __future__ import annotations
from importlib.metadata import version, PackageNotFoundError
from typing import TYPE_CHECKING
import sqlite3
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def get_geoprob_pipe_version_number() -> str:
    geoprob_pipe_version_number = "DEV"
    try: geoprob_pipe_version_number = version('geoprob_pipe')
    except PackageNotFoundError: pass
    return geoprob_pipe_version_number


def get_geohydrological_model(app_settings: ApplicationSettings) -> str:
    # Get geohydrological model
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("""
        SELECT geoprob_pipe_metadata."values" 
        FROM geoprob_pipe_metadata 
        WHERE metadata_type='geohydrologisch_model';
    """)
    result = cursor.fetchone()
    if not result:
        raise ValueError
    model_string = result[0]
    conn.close()

    return model_string




# from pathlib import Path
# from typing import Tuple
# import pandas as pd
# from geoprob_pipe.utils.file_system import FileSystem
# from datetime import datetime
# import logging


# class Workspace:
#     """Workspace class which handles all actions related to input, output and intermediate working files"""
#
#     def __init__(self, path_workspace: str | Path) -> None:
#         """Initialize Workspace instance
#
#         Args:
#             path_workspace (str | Path): path to the folder that contains all required input and where all output and
#                 working files will be stored
#         """
#         self.folderpath = FileSystem.validate_path(path_workspace)
#         self.path_output_folder = _prepare_output_folder(self.folderpath)
#         self.input, self.path_input_excel, self.path_hrd = _prepare_input_folder(self.folderpath)
#         # TODO Later Could Groot: Add functionality to read existing results (without running prob. calculations again)
#         logger.info("Workspace (I/O folders) successfully processed.")
#
#     def update_output_filesystem(self) -> None:
#         """
#         Update the output subfolder FileSystem instance (to include new calculation results)
#         """
#         self.path_output_folder = _prepare_output_folder(self.folderpath)


# def _prepare_output_folder(path_workspace: Path) -> FileSystem:
#     """ Prepare output subfolder
#
#     Args:
#         path_workspace (Path): path to the folder that contains all required input and where all output and working
#             files will be stored
#
#     Raises:
#         FileNotFoundError: raised if USE_EXISTING_TKX_RESULTS=True but the output subfolder contains no .tkx files
#         FileExistsError: raised if USE_EXISTING_TKX_RESULTS=False but the output subfolder still contains .tkx files
#             from a previous run
#
#     Returns:
#         FileSystem: _description_
#     """
#     timestamp = datetime.now().strftime("%Y-%m-%d_%H%M")
#     export_path = path_workspace / "output" / timestamp
#     export_path.mkdir(exist_ok=False, parents=True)
#     logger.info(f"Output folder was successfully created in workspace folder: output/{timestamp}/.")
#     return FileSystem(export_path)


# def _prepare_input_folder(path_workspace: Path) -> Tuple[FileSystem, Path, Path]:
#
#     """Prepare input subfolder
#
#     Args:
#         path_workspace (Path): path to the folder that contains all required input and where all output and working
#             files will be stored
#
#     Raises:
#         FileNotFoundError: raised if USE_EXISTING_TKX_RESULTS=True but the number of .stix files in the input subfolder
#             don't match the number of .tkx files in the output subfolder
#         FileNotFoundError: raised if USE_EXISTING_TKX_RESULTS=False but the input subfolder contains no .stix files
#         FileNotFoundError: raised if USE_EXISTING_TKX_RESULTS=False but the input subfolder contains no template .tkx
#             file
#
#     Returns:
#         FileSystem: FileSystem instance which creates a convenient overview of the input subfolder and the files within it
#     """
#     # Create input folder if it didn't exist yet
#     if not Path(path_workspace / "input").exists():
#         Path.mkdir(path_workspace / "input", parents=False, exist_ok=False)
#         print(f"INFO: input folder was successfully created in project folder ({path_workspace / 'input'})")
#
#     # Create FileSystem instance for input folder
#     filesystem_input = FileSystem(path_workspace / "input")
#
#     # Make sure the "input" subfolder contains 1 .xlsx file named input.xlsx
#     xlsx_input = [filepath.name
#                   for filepath in FileSystem.find_files_in_dir(filesystem_input.folderpath, "xlsx")]
#
#     if len(xlsx_input) != 1 or xlsx_input[0] != "input.xlsx":
#         raise FileNotFoundError(
#             f"\nInput folder {filesystem_input.folderpath} should contain exactly 1 .xlsx files named 'input.xlsx'\n"
#             f"(Currently found: {xlsx_input})"
#         )
#
#     # Check if input.xlsx has exactly four required sheets (case-sensitive!)
#     expected_sheets = {"Vakken", "Uittredepunten", "Ondergrondscenarios", "Overzicht_parameters", "Settings"}
#     xlsx_input_folderpath = pd.ExcelFile(filesystem_input.folderpath / "input.xlsx")
#
#     if set(xlsx_input_folderpath.sheet_names) != expected_sheets:
#         raise FileNotFoundError(
#             f"\n{filesystem_input.folderpath / 'input.xlsx'} should contain exactly these sheets:\n"
#             f"{', '.join(expected_sheets)}\n"
#             f"Found sheets: {', '.join(xlsx_input_folderpath.sheet_names)}\n"
#             f"Note:\n1) Sheet names are case-sensitive!\n2) The order of sheets does not matter."
#         )
#
#     # Find HRD path
#     hrd_matches = filesystem_input.files.loc[
#         lambda df: df["filename"].str.endswith(".sqlite") &
#                 (df["filename"] != "hlcd.sqlite") &
#                 ~df["filename"].str.endswith(".config.sqlite")
#     ]
#     # The .zip files containing HRD have multiple .sqlite files, but we only want the one that ends with .sqlite (not
#     # ending with .config.sqlite) and is not called 'hlcd.sqlite'
#
#     if len(hrd_matches) != 1:
#         if len(hrd_matches) > 1:
#             filepaths = ', '.join(hrd_matches["filepath"].tolist())
#             raise ValueError(f"Expected exactly one matching HRD .sqlite file, but found {len(hrd_matches)}:\n{filepaths}")
#         else:
#             raise ValueError("One HRD .sqlite file is required in the input folder, but none was found.")
#
#     hrd_path = hrd_matches["filepath"].iloc[0]
#
#     return filesystem_input, filesystem_input.folderpath / "input.xlsx", hrd_path



from __future__ import annotations
from pandas import DataFrame
from typing import TYPE_CHECKING, Optional
from geoprob_pipe.results.construct_dataframes import (
    combine_df_beta_per_limit_state, combine_df_beta_per_scenario_rp, combine_df_beta_per_scenario_cp,
    combine_df_beta_per_scenario_final, calculate_df_beta_per_uittredepunt, construct_df_beta_per_vak)
from geoprob_pipe.results.alphas_and_physical_values import construct_df
import os
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Results:
    """ Subclass to intuitively group the results. """

    def __init__(self, geoprob_pipe: GeoProbPipe):
        self.geoprob_pipe = geoprob_pipe
        self.df_beta_limit_states = combine_df_beta_per_limit_state(geoprob_pipe.calc_results)
        self.df_beta_scenarios_rp = combine_df_beta_per_scenario_rp(geoprob_pipe.calc_results)
        # Scenario calculations as a single Reliability Project.
        self.df_beta_scenarios_cp = combine_df_beta_per_scenario_cp(geoprob_pipe.calc_results)
        # Scenario calculations as a single Combine Project.
        self.df_beta_scenarios_final = combine_df_beta_per_scenario_final(geoprob_pipe.calc_results)
        # Worst result from scenario calculations, either combine project, reliability project or max of limit states.
        self._df_alphas_influence_factors_and_physical_values: Optional[DataFrame] = None
        self.df_beta_uittredepunten = calculate_df_beta_per_uittredepunt(geoprob_pipe=geoprob_pipe, results=self)
        self.df_beta_vakken = construct_df_beta_per_vak(self)

    def df_alphas_influence_factors_and_physical_values(
            self,
            filter_deterministic: bool = True,
            filter_derived: bool = False,
    ) -> DataFrame:

        # Generate if not generated yet
        if self._df_alphas_influence_factors_and_physical_values is None:
            self._df_alphas_influence_factors_and_physical_values = (
                construct_df(self.geoprob_pipe))

        # Filters
        df = self._df_alphas_influence_factors_and_physical_values
        if filter_deterministic:
            df = df[df['distribution_type'] != "deterministic"]
        if filter_derived:
            df = df[df['distribution_type'] != "derived"]

        return df

    @property
    def export_dir(self) -> str:
        path: str = os.path.join(
            self.geoprob_pipe.input_data.app_settings.workspace_dir,
            "exports",
            str(self.geoprob_pipe.input_data.app_settings.datetime_stamp),
            "results")
        os.makedirs(path, exist_ok=True)
        return path

    def export_results(
            self,
            bool_beta_limit_states: bool = True,
            bool_beta_scenarios_rp: bool = True,
            bool_beta_scenarios_cp: bool = True,
            bool_beta_scenarios_final: bool = True,
            bool_alphas_influence_factors_and_physical_values: bool = True,
            bool_beta_uittredepunten: bool = True,
            bool_beta_vakken: bool = True):

        # Results of limit state calculations
        if bool_beta_limit_states:
            df = self.df_beta_limit_states
            # TODO Nu Must Klein: Voor export df_beta_limit_states, kolommen filteren?
            df.to_excel(excel_writer=os.path.join(self.export_dir, "df_beta_limit_states.xlsx"))
            # TODO Nu Should Klein: Sommige resultaten zijn niet converged. Wat doen we daarmee?
            #  Op dit moment worden ze gewoon gebruikt om de scenario-faalkans te berekenen.

        if bool_beta_scenarios_rp:
            df = self.df_beta_scenarios_rp
            df.to_excel(excel_writer=os.path.join(self.export_dir, "df_beta_scenarios_rp.xlsx"))

        if bool_beta_scenarios_cp:
            df = self.df_beta_scenarios_cp
            df.to_excel(excel_writer=os.path.join(self.export_dir, "df_beta_scenarios_cp.xlsx"))

        if bool_beta_scenarios_final:
            df = self.df_beta_scenarios_final
            df.to_excel(excel_writer=os.path.join(self.export_dir, "df_beta_scenarios_final.xlsx"))

        if bool_alphas_influence_factors_and_physical_values:
            df = self.df_alphas_influence_factors_and_physical_values()
            df.to_excel(
                excel_writer=os.path.join(self.export_dir, "df_alphas_influence_factors_and_physical_values.xlsx"))

        if bool_beta_uittredepunten:
            self.df_beta_uittredepunten.to_excel(
                excel_writer=os.path.join(self.export_dir, "df_beta_uittredepunten.xlsx"))

        if bool_beta_vakken:
            self.df_beta_vakken.to_excel(excel_writer=os.path.join(self.export_dir, "df_beta_vakken.xlsx"))



"""
In deze `limit_states` subpackage worden alle grenstoestandsfuncties
opgebouwd. Deze worden als probabilistisch model toegevoegd aan de system
design points en individuele design point berekeningen onder de subpackage
`system_calculations`. De grenstoestandsfuncties zijn weer opgebouwd uit
fysieke sub-functies die in de subpackage `physical_components` staan.
"""



# import os
# import random
# from typing import Union, Literal
# import sys
#
# # rgb(153,204,0)
#
#
# # class Tags:
# #
# #     def __init__(self, main_tag: str):
# #         self.main_tag: str = main_tag
# #
# #     @property
# #     def beta(self):
# #         return
# #
#
#
#
#
# class VisualizeInfo:
#
#     def __init__(self, main_tag: str, beta: Union[float, int, str] = "n.b.", visible: bool = False):
#         self.main_tag: str = main_tag
#         self.beta: Union[float, int, str] = beta
#         self.visible: bool = visible
#
#     @property
#     def svg_color(self) -> str:
#         """ Color value formatted for the SVG-file. """
#         if self.beta > 5.00:
#             return "rgb(154,205,50)"  # Green
#         return "rgb(206,32,41)"  # Red
#
#     @property
#     def svg_beta(self) -> str:
#         """ Beta value formatted for the SVG-file. """
#         if isinstance(self.beta, float) or isinstance(self.beta, int):
#             return f"{round(self.beta, 2):.2f}"
#         return str(self.beta)
#
#     @property
#     def svg_visibility(self):
#         """ Visibility value formatted for the SVG-file. """
#         if self.visible:
#             return "visible"
#         return "hidden"
#
#     def svg_tag(self, option: Literal["beta", "color", "visibility"]) -> str:
#         return f"{self.main_tag}.{option}"
#
#     def svg_tag_incl_brackets(self, option: Literal["beta", "color", "visibility"]) -> str:
#         return f"{{{{ {self.svg_tag(option=option)} }}}}"
#
#
# list_visualize_info = [
#     VisualizeInfo(main_tag=, beta=random.random() * 7 + 2, visible=True),
# ]
#
# dict_visualize_info = {
#     "uplift": VisualizeInfo(main_tag="uplift", beta=random.random() * 7 + 2, visible=True),
#     "heave": VisualizeInfo(main_tag="heave", beta=random.random() * 7 + 2, visible=True),
#     "piping": VisualizeInfo(main_tag="piping", beta=random.random() * 7 + 2, visible=True),
#     "scenarios.1": VisualizeInfo(main_tag="scenarios.1", beta=random.random() * 7 + 2, visible=True),
#     "scenarios.2": VisualizeInfo(main_tag="scenarios.2", beta=random.random() * 7 + 2),
# }
#
#
#
#
#
# def generate_overview_flow_chart_with_betas(scenario, uittredepunt):
#     """ Generates a flow chart that provides an overview what the beta values are per step in the calculation process.
#     It displays it from the given scenario and uittredepunt, until vak- and traject-level. """
#
#
#     tags = {
#         "uplift": VisualInfo(random.random() * 7 + 2, True),
#         "heave": VisualInfo(random.random() * 7 + 2, True),
#         "piping": VisualInfo(random.random() * 7 + 2, True),
#         "scenarios": {
#             1: VisualInfo(random.random() * 7 + 2),
#             2: VisualInfo(random.random() * 7 + 2),
#             3: VisualInfo(random.random() * 7 + 2),
#             4: VisualInfo(random.random() * 7 + 2),
#             5: VisualInfo(random.random() * 7 + 2),
#             6: VisualInfo(random.random() * 7 + 2),
#             7: VisualInfo(random.random() * 7 + 2),
#         },
#         "uittredepunten": {
#             1: VisualInfo(random.random() * 7 + 2),
#             2: VisualInfo(random.random() * 7 + 2),
#             3: VisualInfo(random.random() * 7 + 2),
#             4: VisualInfo(random.random() * 7 + 2),
#             5: VisualInfo(random.random() * 7 + 2),
#             6: VisualInfo(random.random() * 7 + 2),
#             7: VisualInfo(random.random() * 7 + 2),
#         },
#         "vakken": {
#             1: VisualInfo(random.random() * 7 + 2),
#             2: VisualInfo(random.random() * 7 + 2),
#             3: VisualInfo(random.random() * 7 + 2),
#             4: VisualInfo(random.random() * 7 + 2),
#             5: VisualInfo(random.random() * 7 + 2),
#             6: VisualInfo(random.random() * 7 + 2),
#             7: VisualInfo(random.random() * 7 + 2),
#         },
#         "traject": VisualInfo(random.random() * 7 + 2),
#
#     }
#
#     # Apply visual and value to each circle
#     concatenated_tags = []
#     values_to_concatenated_tags = []
#     visibility_to_concatenated_tags = []
#     for key, value in tags.items():
#
#         print(f"{key=}, {value=}")
#
#
#         # Determine tag .-string
#         if isinstance(value, VisualInfo):
#             tag_beta = f"{key}.beta"
#             value_beta = value.beta
#             tag_color = f"{key}.color"
#             value_color = value.color
#
#             concatenated_tags.append(tag_beta)
#             values_to_concatenated_tags.append(value_beta)
#             concatenated_tags.append(tag_color)
#             values_to_concatenated_tags.append(value_color)
#
#         elif isinstance(value, dict):
#             # print(f"{concatenated_tags=}")
#             # print(f"{values_to_concatenated_tags=}")
#             # sys.exit()
#             for key2, value2 in value.items():
#                 # print(f"{key2=}, {value2=}")
#                 value2: VisualInfo
#                 tag_beta = f"{key}.{key2}.beta"
#                 # print(f"{value2=}")
#                 value_beta = value2.beta
#                 tag_color = f"{key}.{key2}.color"
#                 value_color = value2.color
#
#                 concatenated_tags.append(tag_beta)
#                 values_to_concatenated_tags.append(value_beta)
#                 concatenated_tags.append(tag_color)
#                 values_to_concatenated_tags.append(value_color)
#
#         else:
#             raise ValueError
#
#     # Read template
#     svg_text = None
#     with open(r"C:\Users\CP\git_clones\GeoProb-Pipe\GeoProb-PipeV2\GeoProb-Pipe\geoprob_pipe\graphs\overview\Hierarchie_berekeningen_incl_result_tags.svg",
#               "r", encoding="utf-8") as f:
#         svg_text = f.read()
#     if svg_text is None:
#         raise ValueError
#
#     # Replace tags
#     for tag, value in zip(concatenated_tags, values_to_concatenated_tags):
#
#         # Prepare tag to search
#         tag_with_brackets = f"{{{{ {tag} }}}}"
#         print(f"{tag_with_brackets=}", tag_with_brackets in svg_text)
#
#         # Prepare value to add
#         if isinstance(value, float) or isinstance(value, int):
#             value_prepped = f"{round(value, 2):.2f}"
#         else:
#             value_prepped = str(value)
#
#         # Add value and color
#         svg_text = svg_text.replace(tag_with_brackets, value_prepped)
#
#     # Save new svg
#     print(f"{svg_text=}")
#     export_dir = r"C:\Users\CP\git_clones\GeoProb-Pipe\GeoProb-PipeV2\exports"
#     with open(os.path.join(export_dir, "updated5.svg"), "w", encoding="utf-8") as f:
#         f.write(svg_text)
#
#     return concatenated_tags, values_to_concatenated_tags
#
#
# generate_overview_flow_chart_with_betas(1, 2)



from __future__ import annotations
from typing import TYPE_CHECKING
from InquirerPy import inquirer
import sys
from geoprob_pipe.utils.validation_messages import BColors

import os
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def added_ahn(app_settings: ApplicationSettings, display_added_msg: bool = False) -> bool:
    # TODO Later Should Middel: Should validate if AHN grid fully overlaps area
    exists = os.path.exists(app_settings.ahn_filepath)

    if exists and display_added_msg:
        print(BColors.OKBLUE, f"✔  AHN-grid al toegevoegd.", BColors.ENDC)
        return True
    if exists and not display_added_msg:
        return True

    return False


def request_ahn(app_settings: ApplicationSettings):

    choices_list = ["Ik heb het nu toegevoegd", "Applicatie afsluiten"]
    choice = inquirer.select(
        message=f"Voeg s.v.p. het AHN-grid toe aan de onderstaande map en met deze bestandsnaam. \n"
                f"{app_settings.ahn_filepath}",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:

        while not os.path.exists(app_settings.ahn_filepath):

            inquirer.select(
                message=f"Voeg s.v.p. het AHN-grid toe aan de map.",
                choices=choices_list,
                default=choices_list[0],
            ).execute()
        print(BColors.OKBLUE, f"✅  AHN-bestand toegevoegd.", BColors.ENDC)
        return True

    elif choice == choices_list[1]:
        sys.exit(f"Applicatie is nu afgesloten.")
    else:
        raise ValueError



from __future__ import annotations
from InquirerPy import inquirer
from pathlib import Path
from typing import TYPE_CHECKING
import scipy.stats as sct
from geopandas import GeoDataFrame, read_file
from shapely import Point
import os
import sqlite3
from geoprob_pipe.utils.validation_messages import BColors
import fiona
import warnings
import time
import pydra_core as pydra
from pandas import DataFrame, concat
from typing import List
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def folder_contains_hrd_db(app_settings: ApplicationSettings) -> bool:
    cnt_sql_files = 0
    cnt_config_files = 0
    cnt_hlcd_files = 0


    for file in os.listdir(app_settings.hrd_dir):
        filename = os.fsdecode(file)
        if filename.endswith(".sqlite"):
            cnt_sql_files += 1
        if filename.endswith(".config.sqlite"):
            cnt_config_files += 1
        if filename.endswith("hlcd.sqlite"):
            cnt_hlcd_files += 1

    if cnt_sql_files == 3 and cnt_config_files == 1 and cnt_hlcd_files == 1:
        return True
    return False


def added_hrd(app_settings: ApplicationSettings) -> bool:
    hrd_files_are_provided: bool = folder_contains_hrd_db(app_settings=app_settings)
    if hrd_files_are_provided:
        print(BColors.OKBLUE, f"✔  HRD-bestanden al toegevoegd.", BColors.ENDC)
        check_hrd_locations_added_to_geopackage(app_settings=app_settings)
        return True

    # Verzoek toe te voegen
    choices_list = ["Ik heb ze nu toegevoegd", "Applicatie afsluiten"]
    choice = inquirer.select(
        message=f"Voeg s.v.p. de bestanden van de hydraulische database toe aan de onderstaande map. Het gaat om alle "
                f"drie SQLite-bestanden, inclusief hlcd.sqlite.\n"
                f"{app_settings.hrd_dir}",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:

        while folder_contains_hrd_db(app_settings=app_settings) is not True:

            inquirer.select(
                message=f"De HRD-bestanden zijn nog niet gevonden in de map. Voeg ze s.v.p. toe.",
                choices=choices_list,
                default=choices_list[0],
            ).execute()
        print(BColors.OKBLUE, f"✅  HRD-bestanden toegevoegd.", BColors.ENDC)
        check_hrd_locations_added_to_geopackage(app_settings=app_settings)
        return True

    elif choice == choices_list[1]:
        return False
    else:
        raise ValueError


def check_hrd_locations_added_to_geopackage(app_settings: ApplicationSettings):

    # Check if already added
    layers = fiona.listlayers(app_settings.geopackage_filepath)
    if "hrd_locaties" in layers:
        print(BColors.OKBLUE, f"✔  HRD-locatie punten al uitgelezen.", BColors.ENDC)
        check_hrd_frag_lines_added_to_geopackage(app_settings=app_settings)
        return

    # Add HRD locations to GeoPackage
    hrd_path = app_settings.hrd_file_path
    hrd = pydra.HRDatabase(hrd_path)
    location_names = hrd.locationnames
    hrd_location_rows = []
    for location_name in location_names:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=FutureWarning)
            try:
                hrd_location = hrd.get_location(location_name)
            except NotImplementedError as e:
                print(f"{BColors.WARNING}Failed adding Hydra-NL location '{location_name}'. "
                      f"Code continues without adding location. "
                      f"Using fragility curve for this location is not possible. "
                      f"Error is: {e}{BColors.ENDC}")
                continue
        hrd_location_rows.append({
            "location_name": location_name,
            "geometry": Point(hrd_location.settings.x_coordinate, hrd_location.settings.y_coordinate)
        })

    # Anticipate no locations added
    # if hrd_location_rows.__len__() == 0:
    #     return

    gdf = GeoDataFrame(hrd_location_rows, columns=['location_name', 'geometry'], crs='EPSG:28992')
    gdf.to_file(Path(app_settings.geopackage_filepath), layer="hrd_locaties", driver="GPKG")
    print(BColors.OKBLUE, f"✅  HRD-locatie punten toegevoegd aan GeoProb-Pipe GeoPackage.", BColors.ENDC)

    check_hrd_frag_lines_added_to_geopackage(app_settings=app_settings)


def check_hrd_frag_lines_added_to_geopackage(app_settings: ApplicationSettings):

    conn = sqlite3.connect(app_settings.geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables_names = [row[0] for row in cursor.fetchall()]
    conn.close()

    # If already exist
    if "fragility_values_invoer_hrd" in tables_names:
        print(BColors.OKBLUE, f"✔  HRD-fragility lines al uitgelezen.", BColors.ENDC)
        return

    # Add frag lines to geopackage
    print(f"{BColors.UNDERLINE}HRD-fragility lines worden nu toegevoegd aan de GeoProb-Pipe GeoPackage.{BColors.ENDC}")
    hrd = pydra.HRDatabase(app_settings.hrd_file_path)
    gdf_locations: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="hrd_locaties")
    location_names = gdf_locations['location_name'].unique().tolist()
    # location_names = hrd.get_location_names()
    fl = pydra.ExceedanceFrequencyLine("h")
    dfs: List[DataFrame] = []
    start_time = time.time()
    last_report = start_time

    for index, location_name in enumerate(location_names):

        # Status report
        if time.time() - last_report >= 10.0:
            print(f"Bezig met locatie {index+1} ({location_name}) van in totaal {location_names.__len__()} locaties.")
            last_report = time.time()

        # TODO:
        #  - Dit proces kan vrij lang duren. Daarom is het beter om per locatie de fragility values in de GeoPackage
        #    te zetten. Dan kan de gebruiker tussentijds afsluiten en later vanaf hetzelfde moment weer oppakken,
        #    indien gewenst.
        #  - Nice to have: In status bericht tijdsindicatie geven wanneer klaar.

        # Continue collecting fragility values
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            location = hrd.get_location(location_name)
        frequency_line = fl.calculate(location)
        df_to_append = DataFrame({
            "fragility_values_ref": [location_name] * frequency_line.level.__len__(),
            "waarde": frequency_line.level,
            "kans": frequency_line.exceedance_frequency,
            "beta": -sct.norm.ppf(frequency_line.exceedance_frequency)
        })
        df_to_append = df_to_append[df_to_append["kans"] < 1.0]
        df_to_append = df_to_append[df_to_append["beta"] < 8.0]
        df_to_append = df_to_append.drop(columns=["beta"])
        dfs.append(df_to_append)

    # Combine data and push
    df = DataFrame(data=[], columns=["fragility_values_ref", "waarde", "kans"])
    if dfs.__len__() > 0:
        df = concat(dfs, ignore_index=True)
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    df.to_sql("fragility_values_invoer_hrd", conn, if_exists="replace", index=False)
    conn.close()

    print(BColors.OKBLUE, f"✅  HRD-fragility lines toegevoegd aan GeoProb-Pipe GeoPackage.", BColors.ENDC)



from __future__ import annotations
import os
import plotly.graph_objects as go
import geopandas as gpd
from shapely.geometry import LineString, MultiLineString, GeometryCollection
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.comparisons import ComparisonCollector


def _determine_zoom(gdf_latlon):
    min_lat = gdf_latlon.geometry.y.min()
    max_lat = gdf_latlon.geometry.y.max()
    min_lon = gdf_latlon.geometry.x.min()
    max_lon = gdf_latlon.geometry.x.max()

    def _calculate_zoom(lat_range_val, lon_range_val):
        max_range = max(lat_range_val, lon_range_val)
        if max_range < 0.01:
            return 15
        elif max_range < 0.05:
            return 13
        elif max_range < 0.1:
            return 12
        elif max_range < 0.5:
            return 10
        elif max_range < 1.0:
            return 9
        else:
            return 8

    lat_range = max_lat - min_lat
    lon_range = max_lon - min_lon
    zoom = _calculate_zoom(lat_range, lon_range)
    return zoom


def _add_line(comparison: ComparisonCollector, fig: go.Figure,
              layer: str, color: str):
    gdf_traject = gpd.read_file(comparison.geopackage_filepath_1,
                                layer=layer)
    gdf_traject = gdf_traject.to_crs("EPSG:4326")

    def plot_linestring(ls, display):
        xs, ys = ls.xy
        xs = list(xs)
        ys = list(ys)
        fig.add_trace(go.Scattermap(
            lon=xs,
            lat=ys,
            mode="lines",
            line=dict(color=color, width=1.5),
            hoverinfo="none",
            name=layer,
            legendgroup=layer,
            showlegend=display
        ))
    show = True
    for geom in gdf_traject.geometry:
        if isinstance(geom, LineString):
            plot_linestring(geom, show)
            show = False
        elif isinstance(geom, MultiLineString):
            for line in geom.geoms:
                plot_linestring(line, show)
                show = False
        elif isinstance(geom, GeometryCollection):
            for g in geom.geoms:
                if isinstance(g, LineString):
                    plot_linestring(g, show)
                    show = False
                elif isinstance(g, MultiLineString):
                    for line in g.geoms:
                        plot_linestring(line, show)
                        show = False

        else:
            print("Skipping unsupported geometry:", geom.geom_type)

    return fig


def map_delta_beta_comparison(comparison: ComparisonCollector,
                              export: bool = False) -> go.Figure:
    # load data from class
    gdf_result1 = (comparison.gdf1_uittredepunten[
        ["uittredepunt_id", "beta", "geometry"]
        ].rename(columns={"beta": "beta1"}))
    gdf_result2 = (comparison.gdf2_uittredepunten[
        ["uittredepunt_id", "beta"]].rename(columns={"beta": "beta2"}))
    gdf = gdf_result1.merge(gdf_result2, on="uittredepunt_id")
    gdf["beta_delta"] = gdf["beta2"] - gdf["beta1"]

    # Covert to crs for map.
    gdf_latlon = gdf.to_crs("EPSG:4326")

    fig = go.Figure()
    fig.add_trace(go.Scattermap(
        mode="markers",
        lat=gdf_latlon.geometry.y,
        lon=gdf_latlon.geometry.x,
        marker=dict(
            size=9,
            color="black"
            ),
        showlegend=False
        ))
    hoverdata = ["uittredepunt_id", "beta_delta", "beta1", "beta2"]
    fig.add_trace(go.Scattermap(
        mode="markers",
        lat=gdf_latlon.geometry.y,
        lon=gdf_latlon.geometry.x,
        marker=dict(
            size=8,
            color=gdf_latlon["beta_delta"],
            cmax=3,
            cmin=-3,
            colorscale="RdYlGn",
            colorbar=dict(title="Delta Beta")
        ),
        hoverinfo='text',
        text=gdf_latlon[hoverdata].apply(
            lambda row: '<br>'.join(
                [f"{col}: {round(row[col], 3)}" for col in hoverdata]
                ),
            axis=1),
        showlegend=False
        ))

    fig = _add_line(comparison, fig, "dijktraject", "black")
    fig = _add_line(comparison, fig, "intredelijn", "blue")
    fig = _add_line(comparison, fig, "binnenteenlijn", "red")
    fig = _add_line(comparison, fig, "buitenteenlijn", "red")

    zoom = _determine_zoom(gdf_latlon)
    fig.update_layout(
        map_style="open-street-map",
        # carto-positron, open-street-map, satellite-streets
        map_zoom=zoom,
        map_center=dict(
            lat=gdf_latlon.geometry.y.mean(),
            lon=gdf_latlon.geometry.x.mean()
        ),
        dragmode="zoom",
        title=f"Delta beta van uittredepunten tussen<br>" +
              f"<sup>Beta1: {comparison.name_1} en Beta2: {comparison.name_2}</sup>",
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="right",
            x=1))

    if export:
        os.makedirs(comparison.export_dir, exist_ok=True)
        fig.write_html(os.path.join(
            comparison.export_dir, "delta_beta_map.html"
            ), include_plotlyjs='cdn', include_mathjax='cdn')
        # fig.write_image(os.path.join(
        #     comparison.export_dir, "delta_beta_map.png"
        #     ), format="png", scale=5,  width=1400)

    return fig


def map_ratio_beta_comparison(comparison: ComparisonCollector,
                              export: bool = False) -> go.Figure:
    # load data from class
    gdf_result1 = (comparison.gdf1_uittredepunten[
        ["uittredepunt_id", "beta", "geometry"]
        ].rename(columns={"beta": "beta1"}))
    gdf_result2 = (comparison.gdf2_uittredepunten[
        ["uittredepunt_id", "beta"]].rename(columns={"beta": "beta2"}))
    gdf = gdf_result1.merge(gdf_result2, on="uittredepunt_id")

    gdf["beta_ratio"] = round((gdf["beta1"] / gdf["beta2"]) * 100, 2)

    # Covert to crs for map.
    gdf_latlon = gdf.to_crs("EPSG:4326")

    fig = go.Figure()
    fig.add_trace(go.Scattermap(
        mode="markers",
        lat=gdf_latlon.geometry.y,
        lon=gdf_latlon.geometry.x,
        marker=dict(
            size=9,
            color="black"
            ),
        showlegend=False
        ))
    hoverdata = ["uittredepunt_id", "beta_ratio", "beta1", "beta2"]
    fig.add_trace(go.Scattermap(
        mode="markers",
        lat=gdf_latlon.geometry.y,
        lon=gdf_latlon.geometry.x,
        marker=dict(
            size=8,
            color=gdf_latlon["beta_ratio"],
            cmax=200,
            cmin=0,
            colorscale="RdYlGn",
            colorbar=dict(title="Beta Ratio - Beta1/Beta2 [%]"),
        ),
        hoverinfo='text',
        text=gdf_latlon[hoverdata].apply(
            lambda row: '<br>'.join(
                [f"{col}: {round(row[col], 3)}" for col in hoverdata]
                ),
            axis=1),
        showlegend=False
        ))

    fig = _add_line(comparison, fig, "dijktraject", "black")
    fig = _add_line(comparison, fig, "intredelijn", "blue")
    fig = _add_line(comparison, fig, "binnenteenlijn", "red")
    fig = _add_line(comparison, fig, "buitenteenlijn", "red")

    zoom = _determine_zoom(gdf_latlon)
    fig.update_layout(
        map_style="open-street-map",
        # carto-positron, open-street-map, satellite-streets
        map_zoom=zoom,
        map_center=dict(
            lat=gdf_latlon.geometry.y.mean(),
            lon=gdf_latlon.geometry.x.mean()
        ),
        dragmode="zoom",
        title=f"Beta ratio van uittredepunten tussen<br>" +
              f"<sup>Beta1: {comparison.name_1} en Beta2: {comparison.name_2}</sup>",
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="right",
            x=1
            )
        )
    if export:
        os.makedirs(comparison.export_dir, exist_ok=True)
        fig.write_html(os.path.join(
            comparison.export_dir, "ratio_beta_map.html"
            ), include_plotlyjs='cdn', include_mathjax='cdn')
        # fig.write_image(os.path.join(
        #     comparison.export_dir, "ratio_beta_map.png"
        #     ), format="png", scale=5,  width=1400)

    return fig



from geoprob_pipe.calculations.systems.moria.system_builder import MoriaSystemBuilder
from geoprob_pipe.calculations.systems.model4a.system_builder import Model4aSystemBuilder
from geoprob_pipe.calculations.systems.wbi.system_builder import WBISystemBuilder
from geoprob_pipe.calculations.limit_states.piping_lm import limit_state_moria
from geoprob_pipe.calculations.systems.model4a.limit_state_functions import limit_state_model4a
# TODO: Dynamisch maken? Forceren dat naamgeving overeenkomt en
# we dynamisch importeren?


CALCULATION_MAPPER = {
    "model4a": {
        "label": "Model 4a",
        "system_builder": Model4aSystemBuilder,
        "system_return_parameter_keys": [
            "z_u", "z_h", "z_p", "z_combin", "h_exit", "r_exit", "phi_exit",
            "d_deklaag", "dphi_c_u", "i_exit","L_voorland", "lambda_voorland",
            "W_voorland", "L_kwelweg", "dh_c", "dh_red"],
        "limit_state_function": limit_state_model4a,
    },
    "wbi": {
        "label": "WBI",
        "system_builder": WBISystemBuilder,
    },
    "moria": {
        "label": "MORIA",
        "system_builder": MoriaSystemBuilder,
        "system_return_parameter_keys": [
            "z_u", "z_h", "z_p", "z_combin", "h_exit", "phi_exit",
            "d_deklaag", "dphi_c_u", "i_exit", "L_voorland", "W_voorland",
            "L_kwelweg", "kD_wvp", "dh_c", "dh_red"],
        "limit_state_function": limit_state_moria,
    },
}



from __future__ import annotations
from InquirerPy import inquirer
import warnings
from typing import Optional, TYPE_CHECKING
import importlib.resources
import os
from geopandas import GeoDataFrame, read_file
import fiona
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def added_dijktraject(app_settings: ApplicationSettings) -> bool:
    layers = fiona.listlayers(app_settings.geopackage_filepath)

    if "dijktraject" in layers:
        print(BColors.OKBLUE, f"✔  Dijktraject al toegevoegd.", BColors.ENDC)
        return True
    else:
        question_trajectory_source(app_settings)
        # TODO Later Should Middel: We vragen nu filepath, we kunnen daarnaast de optie geven voor normtrajecten direct.
        return True


def question_trajectory_source(app_settings: ApplicationSettings):
    choices_list = ["Waterveiligheidsportaal (primaire keringen)", "Lokaal GIS bestand (geopackage/shapefile/geodatabase)"]

    choice = inquirer.select(
        message="Van waaruit wil je de referentielijn van de dijk inladen?",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        with importlib.resources.path(
                package='geoprob_pipe.input_data.dijktrajecten',
                resource='dijktrajecten.shp') as shp_path:
            gdf: GeoDataFrame = read_file(shp_path)
        specify_single_trajectory(
            app_settings, gdf=gdf, column_name="TRAJECT_ID")
    elif choice == choices_list[1]:
        request_trajectory_filepath(app_settings)
    return False


def request_trajectory_filepath(app_settings: ApplicationSettings):
    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar de geopackage/shapefile/geodatabase waarin de "
                    "referentielijn van de dijk zit.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not (filepath.endswith(".gpkg") or filepath.endswith(".shp") or filepath.endswith(".gdb")):
            print(BColors.WARNING, f"Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                                   f"eindigt op de extensie .{filepath.split(sep='.')[-1]}.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    if filepath.endswith(".gdb"):
        specify_geodatabase_layer(app_settings, filepath)
    elif filepath.endswith(".gpkg"):
        specify_geopackage_layer(app_settings, filepath)
    elif filepath.endswith(".shp"):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
            gdf: GeoDataFrame = read_file(filepath)
        specify_column_with_trajectory_name(app_settings, gdf=gdf)
    else:
        raise NotImplementedError(f"File with extension {filepath.split(sep='.')[-1]} is not yet supported. "
                                  f"Please make a request.")


def specify_geodatabase_layer(app_settings: ApplicationSettings, filepath: str):
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de layer waarin met de referentielijn van het dijktraject. Type 'listlayers' om "
                    "een overzicht te krijgen van de geodatabase-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geodatabase: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De layer name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geodatabase: {layers_str}", BColors.ENDC)
            continue
        # TODO Later Must Klein: Check dat een LineString-laag wordt opgegeven.

        layer_name_is_valid = True

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
        gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    specify_column_with_trajectory_name(app_settings, gdf)


def specify_geopackage_layer(app_settings: ApplicationSettings, filepath: str):
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de layer waarin met de referentielijn van het dijktraject. Type 'listlayers' om "
                    "een overzicht te krijgen van de geopackage-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geopackage: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De layer name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geopackage: {layers_str}", BColors.ENDC)
            continue
        # TODO Later Must Klein: Check dat een LineString-laag wordt opgegeven.

        layer_name_is_valid = True

    gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    specify_column_with_trajectory_name(app_settings, gdf)


def specify_column_with_trajectory_name(app_settings: ApplicationSettings, gdf: GeoDataFrame):
    column_name: Optional[str] = None
    column_name_is_valid = False
    while column_name_is_valid is False:
        column_name: str = inquirer.text(
            message="Specificeer de kolom waarin de naam van het dijktraject staat. Type 'listcolumns' om "
                    "een overzicht te krijgen van de kolommen. ",
        ).execute()

        column_names = gdf.columns
        columns_str = ", ".join(column_names)
        if column_name == "listcolumns":
            print(BColors.OKBLUE,
                  f"De volgende kolommen zijn beschikbaar in de spatial layer: {columns_str}", BColors.ENDC)
            continue
        elif column_name not in column_names:
            print(BColors.OKBLUE, f"De kolom naam '{column_name}' bestaat niet. De volgende kolommen zijn beschikbaar "
                                  f"in de spatial layer: {columns_str}", BColors.ENDC)
            continue

        column_name_is_valid = True

    column_name: str
    specify_single_trajectory(app_settings, gdf, column_name)


def specify_single_trajectory(app_settings: ApplicationSettings, gdf: GeoDataFrame, column_name: str):

    # Single item in gdf? Then we know enough
    if gdf.__len__() == 1:
        gdf = gdf[[column_name, gdf.geometry.name]]
        gdf = gdf.rename(columns={
            column_name: "traject_naam",
            gdf.geometry.name: "geometry",
        })
        gdf.to_file(app_settings.geopackage_filepath, layer="dijktraject", driver="GPKG")

        print(BColors.OKBLUE, f"✅  Trajectlijn toegevoegd.", BColors.ENDC)
        return

    # Multiple items in gdf? Make user select single one
    trajectory_name: Optional[str] = None
    trajectory_name_is_valid = False
    while trajectory_name_is_valid is False:
        trajectory_name: str = inquirer.text(
            message=f"Er zijn {gdf.__len__()} opties. Type hier welke de juiste referentielijn is. Type "
                    f"'listoptions' om een overzicht te krijgen van de opties.",
        ).execute()

        trajectory_names = gdf[column_name].values.tolist()
        trajectory_names.sort()
        trajectories_str = ", ".join(trajectory_names)
        if trajectory_name == "listoptions":
            print(BColors.OKBLUE,
                  f"De volgende opties zijn beschikbaar: {trajectories_str}", BColors.ENDC)
            continue
        elif trajectory_name not in trajectory_names:
            print(BColors.OKBLUE, f"De keuze '{trajectory_name}' bestaat niet. De volgende opties zijn beschikbaar: "
                                  f"{trajectories_str}", BColors.ENDC)
            continue

        trajectory_name_is_valid = True

    gdf = gdf[gdf[column_name] == trajectory_name]
    gdf = gdf[[column_name, gdf.geometry.name]]
    gdf = gdf.rename(columns={
        column_name: "traject_naam",
        gdf.geometry.name: "geometry",
    })
    gdf.to_file(app_settings.geopackage_filepath, layer="dijktraject", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Trajectlijn toegevoegd.", BColors.ENDC)
    return



from __future__ import annotations
from geopandas import read_file
from InquirerPy import inquirer
import warnings
from geoprob_pipe.cmd_app.utils.spatial import load_dijktraject_linestring
from geoprob_pipe.utils.gdf import convert_mls_geom_column_to_ls
import os
from pathlib import Path
from shapely import LineString, MultiLineString
from shapely.ops import substring
from typing import TYPE_CHECKING, Optional
from pandas import DataFrame
from geopandas import GeoDataFrame
import fiona
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def added_vakindeling(app_settings: ApplicationSettings) -> bool:
    layers = fiona.listlayers(app_settings.geopackage_filepath)
    if "vakindeling" in layers:
        check_validity_vakindeling(app_settings=app_settings)
        return True
    else:
        request_vakindeling_filepath(app_settings)
        return True


def check_validity_vakindeling(app_settings: ApplicationSettings):
    gdf_dijktraject: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="dijktraject")
    gdf_dijktraject_geom = gdf_dijktraject.iloc[0].geometry
    if isinstance(gdf_dijktraject_geom, MultiLineString):
        assert gdf_dijktraject_geom.geoms.__len__() == 1
        ls_dijktraject: LineString = gdf_dijktraject_geom.geoms[0]
    elif isinstance(gdf_dijktraject_geom, LineString):
        ls_dijktraject = gdf_dijktraject_geom
    else:
        raise NotImplementedError(f"Type of '{type(gdf_dijktraject_geom)} is not yet supported. Please contact the "
                                  f"developer.'")
    dijktraject_length = round(ls_dijktraject.length, 2)

    gdf_vakindeling: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="vakindeling")
    vakindeling_geometries = gdf_vakindeling.geometry.tolist()
    vakindeling_total_length = round(sum([geom.length for geom in vakindeling_geometries]), 2)

    assert dijktraject_length == vakindeling_total_length
    print(BColors.OKBLUE, f"✔  Vakindeling al toegevoegd.", BColors.ENDC)


def import_from_geopackage(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de laag met de vakindeling. "
                    "Type 'listlayers' om een overzicht te krijgen van de geopackage-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geopackage: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geopackage: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
        gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf


def import_from_geodatabase(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de laag met de vakindeling. "
                    "Type 'listlayers' om een overzicht te krijgen van de geodatabase-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geodatabase: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geodatabase: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
        gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf


def request_vakindeling_filepath(app_settings: ApplicationSettings):
    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar de geopackage/shapefile/geodatabase waarin de "
                    "vakindeling van de dijk zit.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not (filepath.endswith(".gpkg") or filepath.endswith(".shp") or filepath.endswith(".gdb")):
            print(BColors.WARNING, f"Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                                   f"eindigt op de extensie .{filepath.split(sep='.')[-1]}.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    if filepath.endswith(".shp"):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
            gdf: GeoDataFrame = read_file(filepath)
        validate_vakindeling(app_settings, gdf=gdf)
    elif filepath.endswith(".gpkg"):
        gdf: GeoDataFrame = import_from_geopackage(filepath=filepath)
        validate_vakindeling(app_settings, gdf=gdf)
    elif filepath.endswith(".gdb"):
        gdf: GeoDataFrame = import_from_geodatabase(filepath=filepath)
        validate_vakindeling(app_settings, gdf=gdf)
    else:
        raise NotImplementedError(f"File with extension {filepath.split(sep='.')[-1]} is not yet supported. "
                                  f"Please make a request.")


def validate_vakindeling(app_settings: ApplicationSettings, gdf: GeoDataFrame):
    """ Validates the vakindeling shape, with some conversions if they can be
    applied safely. """
    gdf = convert_mls_geom_column_to_ls(gdf=gdf)
    assert gdf.geometry.apply(lambda geom: isinstance(geom, LineString)).all(), \
        ("De opgegeven vakindeling heeft niet voor elk vak een geometry. De "
         "applicatie sluit nu af.")
    specify_column_with_vaknaam(app_settings, gdf=gdf)


def specify_column_with_vaknaam(
        app_settings: ApplicationSettings, gdf: GeoDataFrame):
    column_name: Optional[str] = None
    column_name_is_valid = False
    while column_name_is_valid is False:
        column_name: str = inquirer.text(
            message="Specificeer de kolom waarin de vaknaam staat. Type "
                    "'listcolumns' om een overzicht te krijgen van de "
                    "kolommen. ",
        ).execute()

        column_names = gdf.columns
        columns_str = ", ".join(column_names)
        if column_name == "listcolumns":
            print(BColors.OKBLUE,
                  f"De volgende kolommen zijn beschikbaar in de spatial "
                  f"layer: {columns_str}", BColors.ENDC)
            continue
        elif column_name not in column_names:
            print(BColors.OKBLUE,
                  f"De kolom naam '{column_name}' bestaat niet. De volgende "
                  f"kolommen zijn beschikbaar in de spatial layer: "
                  f"{columns_str}", BColors.ENDC)
            continue

        column_name_is_valid = True

    column_name: str
    specify_column_with_vak_id(
        app_settings, gdf=gdf, kolom_vak_naam=column_name)


def is_numeric_integer(val):
    try:
        return float(val) % 1 == 0
    except (ValueError, TypeError):
        return False


def specify_column_with_vak_id(
        app_settings: ApplicationSettings, gdf: GeoDataFrame,
        kolom_vak_naam: str):
    kolom_vak_id: Optional[str] = None
    column_name_is_valid = False
    while column_name_is_valid is False:
        kolom_vak_id: str = inquirer.text(
            message="Specificeer de kolom waarin het vak id staat. Indien "
                    "onnodig, type 'nvt'. Type 'listcolumns' om een overzicht "
                    "te krijgen van de kolommen. ",
        ).execute()

        column_names = gdf.columns
        columns_str = ", ".join(column_names)
        if kolom_vak_id.lower() == "nvt":
            align_vak_shp_to_dijktraject(
                app_settings, gdf_vakindeling=gdf,
                kolom_vak_naam=kolom_vak_naam, kolom_vak_id=None)
            return
        elif kolom_vak_id == "listcolumns":
            print(BColors.OKBLUE,
                  f"De volgende kolommen zijn beschikbaar in de spatial "
                  f"layer: {columns_str}", BColors.ENDC)
            continue
        elif kolom_vak_id not in column_names:
            print(f"{BColors.OKBLUE}De kolom naam '{kolom_vak_id}' bestaat "
                  f"niet. De volgende kolommen zijn beschikbaar in de spatial "
                  f"layer: {columns_str}{BColors.ENDC}")
            continue

        # Ensure column values are unique and integers
        if gdf[kolom_vak_id].__len__() != gdf[kolom_vak_id].unique().__len__():
            print(f"{BColors.OKBLUE}De waarden in deze kolom zijn niet uniek. "
                  f"Corrigeer de dubbelingen, of kies een andere kolom."
                  f"{BColors.ENDC}")
            continue

        elif not gdf[kolom_vak_id].apply(is_numeric_integer).all():
            print(f"{BColors.OKBLUE}De waarden in deze kolom zijn niet allen "
                  f"volledige getallen (integers). Corrigeer de kolom, of "
                  f"kies een andere.{BColors.ENDC}")
            continue

        column_name_is_valid = True

    kolom_vak_id: str
    align_vak_shp_to_dijktraject(
        app_settings, gdf_vakindeling=gdf, kolom_vak_naam=kolom_vak_naam,
        kolom_vak_id=kolom_vak_id)


def align_vak_shp_to_dijktraject(
        app_settings: ApplicationSettings, gdf_vakindeling: GeoDataFrame ,
        kolom_vak_naam: str, kolom_vak_id: Optional[str] = None):

    # Get dijktraject linestring
    ls_dijktraject = load_dijktraject_linestring(app_settings=app_settings)

    # Data verzamelen uit provided vak shp
    rows = []
    for index, row in gdf_vakindeling.iterrows():
        pnt1 = row.geometry.boundary.geoms[0]
        pnt2 = row.geometry.boundary.geoms[1]
        m_pnt1 = round(ls_dijktraject.project(pnt1), 1)
        m_pnt2 = round(ls_dijktraject.project(pnt2), 1)
        m_start = min(m_pnt1, m_pnt2)
        new_row = {"naam": row[kolom_vak_naam], "m_start": m_start}
        if kolom_vak_id:
            new_row['id'] = int(row[kolom_vak_id])
        rows.append(new_row)
    df = DataFrame(rows)
    df = df.sort_values(by=["m_start"], ignore_index=True)
    df['m_end'] = df['m_start'].shift(periods=-1)

    # Add zero and final dijktraject length to first and last vak
    df.loc[0, 'm_start'] = 0
    df.loc[max(df.index), 'm_end'] = ls_dijktraject.length

    # Align vak geometry to dijktraject by retrieving substring
    rows = []
    for index, row in df.iterrows():
        new_row = {
            "naam": row["naam"],
            "m_start": row["m_start"],
            "m_end": row["m_end"],
            "geometry": substring(ls_dijktraject, row["m_start"], row["m_end"])
        }
        if kolom_vak_id:
            new_row["id"] = row['id']
        rows.append(new_row)
    gdf_new_vakindeling = GeoDataFrame(rows, crs='EPSG:28992')
    if not kolom_vak_id:
        gdf_new_vakindeling['id'] = gdf_new_vakindeling.index + 1
    gdf_new_vakindeling: GeoDataFrame = gdf_new_vakindeling[
        ["id", "naam", "m_start", "m_end", "geometry"]]

    # Add to geopackage
    gdf_new_vakindeling.to_file(
        Path(app_settings.geopackage_filepath),
        layer="vakindeling", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Vakindeling toegevoegd.", BColors.ENDC)



from __future__ import annotations
import os
from datetime import datetime
from typing import TYPE_CHECKING, Optional, List
import pandas as pd

try:
    import probabilistic_library
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "No module named 'probabilistic_library'. This package is not publicly available or part of the repository. \n"
        "Please request the wheel-file through the developer and install it manually. Due to copyright reasons, do \n"
        "not commit the wheel-file into the repository.")

from geoprob_pipe.input_data import InputData
from geoprob_pipe.results import Results
from geoprob_pipe.spatial import Spatial
from geoprob_pipe.visualizations import Visualizations
from geoprob_pipe.calculations.systems.build_and_run import build_and_run_system_calculations
from geoprob_pipe.utils.update_metadata import update_metadata
import logging
if TYPE_CHECKING:
    from geoprob_pipe.calculations.systems.build_and_run import CalcResult
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


logger = logging.getLogger("geoprob-pipe")


class GeoProbPipe:
    """ GeoProb-Pipe application object. """
    # TODO Later Could Groot: Gebruiker optie geven OpenTurns of Prob-library te kiezen? Dus engine keuze.

    def __init__(self, app_settings: ApplicationSettings) -> None:

        # Miscellaneous
        import warnings
        warnings.simplefilter(
            action='ignore',
            category=FutureWarning)  # Preferably address FutureWarnings: part of pydra-core

        logger.info("Initiating project.")
        self.time_start = datetime.now()

        self.input_data = InputData(app_settings=app_settings)  # TODO: Alter with new option

        # Read calculation settings
        # self._read_calculation_settings()  # TODO: Not part of new version
        # TODO: Unsure if the single statement belongs here. Wouldn't it be part of input data?

        self.calc_results: List[CalcResult] = build_and_run_system_calculations(self)
        self.results = Results(self)

        # Log finish
        self.time_end = datetime.now()
        self.time_diff = self.time_end - self.time_start
        logger.info(f"Calculations were performed successfully in {int(self.time_diff.total_seconds())} seconds.")

        

        # Append logic classes
        self.visualizations = Visualizations(self)
        self.spatial = Spatial(self)

    # def _read_calculation_settings(self):  # TODO: Not (yet) part of new version
    #     """ Read calculation settings from Excel file. """
    #     self.df_settings = read_excel(self.workspace.path_input_excel, sheet_name="Settings", index_col=0, header=0)
    #     logger.info(f"Settings successfully loaded from `{self.workspace.path_input_excel.name}`.")
    #     time.sleep(1)  # Some time to make sure the print below, is printed after the logger print.

    def _export_validation_messages(self):
        # Gather validation messages from calculations
        df_val: Optional[pd.DataFrame] = None
        [result.validation_message.concat_with_df(df_to_append_to=df_val) for result in self.calc_results
         if result.validation_message is not None]

        # Export dataframe with validation messages
        if df_val is not None:
            export_path = os.path.join(self.input_data.app_settings.workspace_dir, "validation_messages.xlsx")
            df_val.to_excel(export_path)

    def export_archive(self):
        """ Exports everything related to this project. """
        logger.info("Now exporting archive...")
        self.results.export_results()
        self.visualizations.export_visualizations()
        self.spatial.export_geopackage()
        # add run metadata to geopackage
        update_metadata(self)
        self._export_validation_messages()

        path: str = os.path.join(
            str(self.input_data.app_settings.workspace_dir), "exports",
            str(self.input_data.app_settings.datetime_stamp))
        print(f"Exported archive to {path}")



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Tuple
import os
from pandas import DataFrame, merge
from geopandas import GeoDataFrame, read_file
from plotly.graph_objects import Figure, Bar
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


DISTINCTIVE_COLORS = [
    '(60, 180, 75)', '(230, 25, 75)', '(255, 225, 25)', '(0, 130, 200)', '(245, 130, 48)', '(145, 30, 180)',
    '(70, 240, 240)', '(240, 50, 230)', '(210, 245, 60)', '(250, 190, 212)', '(0, 128, 128)',
    '(220, 190, 255)', '(170, 110, 40)', '(255, 250, 200)', '(128, 0, 0)', '(170, 255, 195)',
    '(128, 128, 0)', '(255, 215, 180)', '(128, 128, 128)', '(255, 255, 255)']


def get_plot_order(geoprob_pipe: GeoProbPipe) -> DataFrame:
    """ Assigns color codes (indices) and color rgb-values to each stochast such that they can be plotted with the
    same color and same order.

    :param geoprob_pipe:
    :return: DataFrame with columns 'variable', 'plot_order' and 'color'
    """

    # Define plot order
    df: DataFrame = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(filter_derived=True)
    df = df[["variable", "influence_factor"]]
    df = df.groupby(['variable']).mean()  # Influence factor average value define color order
    df = df.sort_values(by=["influence_factor"], ascending=False)
    df = df.reset_index(drop=False)
    df['plot_order'] = df.index  # Result is df with columns 'variable', 'influence_factor' (average), 'plot_order'

    # Add colors to plot order
    df_colors = DataFrame({"color": DISTINCTIVE_COLORS})  # Convert to dict to df for ease of use
    df_colors['plot_order'] = df_colors.index
    df = merge(df, df_colors, how="left", on="plot_order")

    return df


def _get_method_used(geoprob_pipe: GeoProbPipe, worst_uittredepunt_id: int, worst_scenario_id: str):
    df_scenario_final: DataFrame = geoprob_pipe.results.df_beta_scenarios_final
    df_scenario_final = df_scenario_final[
        (df_scenario_final["uittredepunt_id"] == worst_uittredepunt_id) &
        (df_scenario_final["ondergrondscenario_id"] == worst_scenario_id)
    ]
    assert df_scenario_final.__len__() == 1

    method_used_str: str = df_scenario_final['method_used'].iloc[0]
    method_used_nr = int(method_used_str[0])

    # If Combine Project or Reliability Project
    if method_used_nr == 1:
        return "CP"
    if method_used_nr == 2:
        return "RP"

    # Else look up which Limit State Triggered
    df_limit_states: DataFrame = geoprob_pipe.results.df_beta_limit_states
    df_limit_states = df_limit_states[
        (df_limit_states["uittredepunt_id"] == worst_uittredepunt_id) &
        (df_limit_states["ondergrondscenario_id"] == worst_scenario_id)
    ]
    assert df_limit_states.__len__() == 3
    df_limit_states = df_limit_states.sort_values(by=["beta"], ascending=False)
    limit_state_used = df_limit_states["limit_state"].iloc[0]
    possible_return_values = {"calc_Z_u": "Zu", "calc_Z_h": "Zh", "calc_Z_p": "Zp"}
    return possible_return_values[limit_state_used]


def get_influence_factors_for_vak(
        geoprob_pipe: GeoProbPipe, df_invloedsfactoren: DataFrame, vak_id: int
) -> Tuple[Optional[DataFrame], int, str, str]:
    """ Returns the influence factors for the worst result of a scenario within the vak. May be empty dataframe if
    vak has no results.

    In the past weigh it for the uittredepunt, or average is among all uittredepunten, as was done in the past. The
    choice was made to keep the actual resulting influence factors because they sum up to 100%. There is also no
    physical reason to average them, and the worst case scenario is normative for the final result anyway.

    :param geoprob_pipe:
    :param df_invloedsfactoren:
    :param vak_id:
    :return: Influence factors, worst case uittredepunt ID, worst case scenario label and method used.
    """
    """  """
    df_result = geoprob_pipe.results.df_beta_scenarios_final
    df_result = df_result[df_result["vak_id"] == vak_id]

    # Check if results for vak
    if df_result.__len__() == 0:
        return None, -1, "", ""

    df_result = df_result.sort_values(by=["beta"], ascending=True)
    worst_uittredepunt_id = df_result.iloc[0]['uittredepunt_id']
    worst_scenario_id = df_result.iloc[0]['ondergrondscenario_id']
    df = df_invloedsfactoren[
        (df_invloedsfactoren["uittredepunt_id"] == worst_uittredepunt_id) &
        (df_invloedsfactoren["ondergrondscenario_id"] == worst_scenario_id)]
    df = df.sort_values(by=["plot_order"])

    # Determine method used
    method_used = _get_method_used(
        geoprob_pipe=geoprob_pipe, worst_uittredepunt_id=worst_uittredepunt_id, worst_scenario_id=worst_scenario_id)
    mapper = {"Zp": "3: Max Limit States (calc_Z_p)", "Zu": "3: Max Limit States (calc_Z_u)",
              "Zh": "3: Max Limit States (calc_Z_h)", "CP": "1: Combine Project", "RP": "2: Reliability Project"}
    df = df[df["design_point"] == mapper[method_used]]

    return df, worst_uittredepunt_id, worst_scenario_id, method_used


def _get_data(geoprob_pipe: GeoProbPipe) -> DataFrame:
    # Get data
    df_invloedsfactoren = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(filter_derived=True)

    # Add plot order
    df_plot_order = get_plot_order(geoprob_pipe=geoprob_pipe)
    df_invloedsfactoren = merge(
        df_invloedsfactoren, df_plot_order[["variable", "plot_order", "color"]], how="left", on="variable")

    return df_invloedsfactoren


def _plot_data(geoprob_pipe: GeoProbPipe, df_invloedsfactoren: DataFrame) -> Figure:
    fig = Figure()
    gdf_vakindeling: GeoDataFrame = read_file(
        geoprob_pipe.input_data.app_settings.geopackage_filepath, layer="vakindeling")
    vakken = {row['id']: row["naam"] for index, row in gdf_vakindeling.iterrows()}
    picked_colors = {}
    added_to_legend = []
    for vak_id, vak_naam in vakken.items():

        # Get invloedsfactoren data
        df_factoren, worst_uittredepunt_id, worst_scenario_id, method_used = get_influence_factors_for_vak(
            geoprob_pipe=geoprob_pipe, df_invloedsfactoren=df_invloedsfactoren, vak_id=vak_id)
        vak_has_results = worst_uittredepunt_id != -1
        x_axis_label = f"{vak_id}"
        if vak_has_results:
            x_axis_label = f"{vak_id} ({worst_uittredepunt_id}, {worst_scenario_id})"

        # If not results/factors
        if df_factoren is None:
            color = f"rgb{DISTINCTIVE_COLORS[0]}"
            # Add dummy zero value to visualize no results in vak
            # noinspection PyTypeChecker
            fig.add_trace(Bar(x=[x_axis_label], y=[0], name="dummy", marker_color=color, showlegend=False))
            continue

        # Plot data
        stochasten = df_factoren['variable'].unique().tolist()
        for index, stochast in enumerate(stochasten):
            if stochast not in picked_colors.keys():
                color = f"rgb{DISTINCTIVE_COLORS[picked_colors.__len__()]}"
                picked_colors[stochast] = color
            else:
                color = picked_colors[stochast]
            show_legend = False
            if stochast not in added_to_legend:
                added_to_legend.append(stochast)
                show_legend = True

            # Plot, including label?
            if index + 1 == stochasten.__len__():
                fig.add_trace(Bar(
                    x=[x_axis_label],
                    y=[df_factoren[df_factoren['variable'] == stochast].iloc[0]['influence_factor'] * 100],
                    text=method_used, textposition='outside',
                    name=stochast, marker_color=color, showlegend=show_legend, legendgroup=stochast))
            else:
                fig.add_trace(Bar(
                    x=[x_axis_label],
                    y=[df_factoren[df_factoren['variable'] == stochast].iloc[0]['influence_factor'] * 100],
                    name=stochast, marker_color=color, showlegend=show_legend, legendgroup=stochast))

    return fig


def _update_layout(fig: Figure) -> Figure:
    fig.update_layout(
        barmode='stack',
        bargap=0,       # Geen ruimte tussen groepen
        bargroupgap=0,  # Geen ruimte tussen individuele bars binnen een groep
        title='Invloedsfactoren<br>'
              '<sup>Invloedsfactoren zijn van de worstcase Beta (combinatie uittredepunt en ondergrondscenario, '
              'zie x-axis label). De labels boven de barchart (CP, RP, Zu, Zh of Zp) zijn een verwijzing naar de '
              'gebruikte design points.</sup>',
        xaxis_title='Vak ID',
        yaxis=dict(title="Percentage", ticksuffix="%"))
    return fig


def invloedsfactoren(geoprob_pipe: GeoProbPipe, export: bool = False) -> Figure:

    # Logic
    df_invloedsfactoren = _get_data(geoprob_pipe=geoprob_pipe)
    fig = _plot_data(geoprob_pipe=geoprob_pipe, df_invloedsfactoren=df_invloedsfactoren)
    fig = _update_layout(fig=fig)

    # Export
    if export:
        export_dir = geoprob_pipe.visualizations.graphs.export_dir
        os.makedirs(export_dir, exist_ok=True)
        fig.write_html(os.path.join(export_dir, f"invloedsfactoren.html"), include_plotlyjs='cdn')

    # for stochast in
    return fig



from __future__ import annotations
from InquirerPy import inquirer
from geoprob_pipe import GeoProbPipe
import sys
from geoprob_pipe.utils.validation_messages import BColors
from typing import TYPE_CHECKING, List, Tuple

if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def _convert_vak_ids_str_to_ints(input_str: str) -> Tuple[List[int], List[str]]:
    input_str = input_str.replace(" ", "")
    list_vak_ids_str = input_str.split(sep=",")

    list_vak_ids_int: List[int] = []
    unknown_inputs: List[str] = []
    for item in list_vak_ids_str:
        if "-" in item:
            try:
                left_bound = int(item.split(sep="-")[0])
            except ValueError:
                unknown_inputs.append(item)
                continue
            try:
                right_bound = int(item.split(sep="-")[1])
            except ValueError:
                unknown_inputs.append(item)
                continue
            ids = list(range(left_bound, right_bound+1))
            list_vak_ids_int.extend(ids)
        else:
            try:
                item_int = int(item)
            except ValueError:
                unknown_inputs.append(item)
                continue
            list_vak_ids_int.append(item_int)

    return list(set(list_vak_ids_int)), unknown_inputs

def request_vakken_to_run() -> str:
    list_vak_ids_int: List[int] = []
    vakken_input_is_valid = False
    while vakken_input_is_valid is False:
        vakken_input: str = inquirer.text(
            message="Specificeer welke vakken je wilt doorrekenen. Doe dit door comma-separated de vak nummers (id) "
                    "op te geven. Bijvoorbeeld '4,5,6,7,10-15'.",
        ).execute()

        # Convert string input to integer values
        list_vak_ids_int, unknown_inputs = _convert_vak_ids_str_to_ints(input_str=vakken_input)
        if unknown_inputs.__len__() > 0:
            print(f"{BColors.WARNING}De applicatie heeft enkele niet valide items in je invoer geconstateerd. "
                  f"Pas deze aan. Je invoer is '{vakken_input}'. De niet valide items zijn '{unknown_inputs}'."
                  f"{BColors.ENDC}")
            continue

        # Assure more than one valid vak id is given
        if list_vak_ids_int.__len__() == 0:
            print(f"{BColors.WARNING}Er zijn geen vak identificatie nummers opgegeven. Je invoer is '{vakken_input}'. "
                  f"Probeer opnieuw.{BColors.ENDC}")
            continue

        vakken_input_is_valid = True

    return f"vakken:{','.join([str(item) for item in list_vak_ids_int])}"


def run_calculations(app_settings: ApplicationSettings) -> bool:
    choices_list = [
        "Ja, alles",
        "Ja, specifieke vakken",
        "Nee (applicatie sluit af)"
    ]
    choice = inquirer.select(
        message="Wil je de berekeningen uitvoeren? \n"
                "Let op: Indien je al rekenresultaten hebt, zullen deze verwijderd worden, alvorens de nieuwe "
                "resultaten worden opgeslagen. ", choices=choices_list, default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        geoprob_pipe = GeoProbPipe(app_settings)
        geoprob_pipe.export_archive()
        print(BColors.OKBLUE, f"✅  Berekeningen zijn uitgevoerd.", BColors.ENDC)
    elif choice == choices_list[1]:
        vakken_str: str = request_vakken_to_run()
        app_settings.to_run = vakken_str
        geoprob_pipe = GeoProbPipe(app_settings)
        geoprob_pipe.export_archive()
        print(BColors.OKBLUE, f"✅  Berekeningen zijn uitgevoerd, voor vakken "
                              f"{app_settings.to_run.replace("vakken:", "")}.", BColors.ENDC)
    elif choice == choices_list[2]:
        sys.exit(f"Applicatie is afgesloten.")
    else:
        raise ValueError
    return True



from typing import Dict, List
from pandas import DataFrame, isna, notna, concat, read_sql, read_csv
import sqlite3
import numpy as np
import os
import scipy.stats as sct
from geopandas import GeoDataFrame, read_file
from geoprob_pipe.calculations.systems.mappers.initial_input_mapper import INITIAL_INPUT_MAPPER
from geoprob_pipe.cmd_app.parameter_input.input_parameter_tables import InputParameterTables
from probabilistic_library import FragilityValue


def _combine_parameter_invoer_sources(tables: InputParameterTables) -> DataFrame:
    """ Combineert de geo-gerefereerde parameter invoer met de handmatige invoer die oorspronkelijk uit de Excel kwam.
    Zodoende kan vanuit één dataframe de invoer geëxplodeerd worden naar invoer per uittredepunt. """

    # Gather raw data
    df_gis_join_parameter_invoer = tables.df_gis_join_parameter_invoer
    df_gis_join_parameter_invoer['scope'] = 'gis_uittredepunt'
    df_parameter_invoer = tables.df_parameter_invoer

    # Concatenate
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        df_parameter_invoer_combined = concat(
            [df_gis_join_parameter_invoer, df_parameter_invoer], ignore_index=True)

    return df_parameter_invoer_combined


def _gather_hrd_frag_line_from_geopackage(ref: str, geopackage_filepath: str):
    # Read database
    conn = sqlite3.connect(geopackage_filepath)
    df_frag_line: DataFrame = read_sql(
        sql=f"SELECT * FROM fragility_values_invoer_hrd WHERE fragility_values_ref = '{ref}' AND kans < 1.0;",
        con=conn)
    conn.close()

    # Filter beta > 8 (probabilistic library cannot work with that in 26.1.1)
    df_frag_line['beta'] = -sct.norm.ppf(df_frag_line['kans'])
    df_frag_line = df_frag_line[df_frag_line['beta'] < 8.0].copy(deep=True)
    df_frag_line = df_frag_line.drop(columns=["beta"])

    # Validate
    assert df_frag_line.__len__() > 2
    # It should be validated beforehand that all added fragility lines are at least 3 points. So if this assert triggers
    # something should be improved earlier in validation.

    # Construct Fragility Values
    df_frag_line = df_frag_line.sort_values(by=["waarde"])
    frag_points = []
    for index, row in df_frag_line.iterrows():
        fc = FragilityValue()
        fc.x = row["waarde"]
        fc.probability_of_failure = row["kans"]
        frag_points.append(fc)

    return frag_points


def _gather_frag_line_from_csv(csv_file_name: str, geopackage_filepath: str):

    # Read csv-file
    csv_dir = os.path.join(os.path.dirname(geopackage_filepath), "frag_csv_files")
    os.makedirs(csv_dir, exist_ok=True)
    path_to_csv = os.path.join(csv_dir, csv_file_name)
    if not os.path.exists(path_to_csv):
        raise FileNotFoundError(
            f"CSV-file with fragility curve not found for reference '{csv_file_name}'. Please make sure to place your "
            f"csv-files at the following location: \n{csv_dir}")
    df_frag_line = read_csv(path_to_csv, sep=",")

    # Validate
    assert df_frag_line.__len__() > 2
    # It should be validated beforehand that all added fragility lines are at least 3 points. So if this assert triggers
    # something should be improved earlier in validation.

    # Construct Fragility Values
    df_frag_line = df_frag_line.sort_values(by=["waarde"])
    frag_points = []
    for index, row in df_frag_line.iterrows():
        fc = FragilityValue()
        fc.x = row["waarde"]
        fc.probability_of_failure = row["kans"]
        frag_points.append(fc)

    return frag_points


def _collect_fragility_values(
        tables: InputParameterTables, fragility_refs: List[str], geopackage_filepath: str,
) -> DataFrame:
    """ Collects the fragility values from the different sources. The sources are (a) the HRD-file, (b) the Excel input
    file, and (c) the csv folder.

    :param tables:
    :param fragility_refs:
    :param geopackage_filepath:
    :return: Returns a dataframe with columns fragility_values_ref and fragility_values.
    """

    df_frag_invoer = tables.df_fragility_values_invoer
    available_frag_invoer_refs = df_frag_invoer['fragility_values_ref'].unique()
    return_array = []
    for fragility_ref in fragility_refs:

        # From .csv-file (not stored in GeoPackage)
        if fragility_ref.endswith(".csv"):
            return_array.append({
                "fragility_values_ref": fragility_ref,
                "fragility_values": _gather_frag_line_from_csv(
                    csv_file_name=fragility_ref, geopackage_filepath=geopackage_filepath)})

        # From HRD-database (previously stored in GeoPackage)
        elif fragility_ref not in available_frag_invoer_refs:
            return_array.append({
                "fragility_values_ref": fragility_ref,
                "fragility_values": _gather_hrd_frag_line_from_geopackage(
                    ref=fragility_ref, geopackage_filepath=geopackage_filepath)})

        # Otherwise, retrieve custom curve from Excel (previously stored in GeoPackage)
        else:
            raise NotImplementedError(f"Should now retrieve it from the df_frag_invoer.")  # TODO

    # Build dataframe
    if return_array.__len__() == 0:
        df = DataFrame(data=[], columns=["fragility_values_ref", "fragility_values"])
        return df
    return DataFrame(return_array)


def _add_fragility_values_to_combined_parameter_invoer(
        df_parameter_invoer_combined: DataFrame, tables: InputParameterTables,
        geopackage_filepath: str, drop_ref: bool = True) -> DataFrame:
    """ Haalt uit de fragility values Excel de arrays op en vervang in de df_parameter_invoer_combined de referentie
    met de daadwerkelijke fragility values. """

    df = df_parameter_invoer_combined.copy(deep=True)

    # Replace empty values with NaN
    df['fragility_values_ref'] = df['fragility_values_ref'].replace('', np.nan)
    df['fragility_values_ref'] = df['fragility_values_ref'].infer_objects(copy=False)

    # Gather referenced fragility value refs
    fragility_refs = df['fragility_values_ref'].dropna().unique()

    # Collect fragility lines
    df_frag_lines = _collect_fragility_values(
        tables=tables, fragility_refs=fragility_refs, geopackage_filepath=geopackage_filepath)

    # Attach to parameter invoer df
    df = df.merge(
        df_frag_lines, left_on="fragility_values_ref", right_on="fragility_values_ref", how="left")

    if drop_ref:
        df = df.drop(columns=["fragility_values_ref"])

    return df


def _collect_right_columns_combined_parameter_invoer(df_parameter_invoer_combined: DataFrame) -> DataFrame:
    """ Parameter tabel omzetten naar juiste kolommen. Enkel per uittredepunt, scenario en parameter de
    parameterinvoer. """

    # Add parameter invoer: op uittredepunten niveau
    df_parameter_invoer_combined = df_parameter_invoer_combined.drop(columns=["bronnen", "opmerking"])
    parameter_description_columns = [
        "distribution_type", "mean", "variation", "deviation", "minimum", "maximum", "fragility_values"]
    if "fragility_values_ref" in df_parameter_invoer_combined.columns:
        parameter_description_columns.append("fragility_values_ref")
    df_parameter_invoer_combined['parameter_input'] = df_parameter_invoer_combined.apply(lambda row: {
        k: row[k] for k in parameter_description_columns if isinstance(row[k], list) or notna(row[k])
    }, axis=1)
    df_parameter_invoer_combined = df_parameter_invoer_combined.drop(columns=parameter_description_columns)
    return df_parameter_invoer_combined


def _construct_df_identifiers(geopackage_filepath: str, tables: InputParameterTables):
    """ Create identifiers table.

    The identifiers table is a table with unique rows (unique uittredepunt, vak and scenario combo). It is used as a
    base to explode to input per scenario and uittredepunt. """

    gdf_uittredepunten: GeoDataFrame = read_file(geopackage_filepath, layer="uittredepunten")
    df_identifiers: DataFrame = gdf_uittredepunten[["uittredepunt_id", "vak_id"]]
    df_scenario_invoer: DataFrame = tables.df_scenario_invoer
    df_identifiers = df_identifiers.merge(df_scenario_invoer, left_on="vak_id", right_on="vak_id")
    df_identifiers = df_identifiers.drop(columns=["kans"])
    return df_identifiers


def _gather_required_input_parameters(geopackage_filepath: str) -> List[str]:

    conn = sqlite3.connect(geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("""
        SELECT geoprob_pipe_metadata."values" 
        FROM geoprob_pipe_metadata 
        WHERE metadata_type='geohydrologisch_model';
    """)
    result = cursor.fetchone()
    if not result:
        raise ValueError
    model_string = result[0]
    conn.close()
    df_dummy_data = DataFrame(INITIAL_INPUT_MAPPER[model_string]['input'])

    _ = df_dummy_data.sort_values(by=["name"])
    df_dummy_data = df_dummy_data.sort_values(by=["name"])
    return df_dummy_data['name'].unique().tolist()
    # TODO: Return this to use in iteration to retrieve data

    # Method to use the system_function. Not necessary because of the dummy input. But for now kept.
    # from geoprob_pipe.calculations.system_calculations.piping_moria.reliability_calculation import \
    #     PipingMORIASystemReliabilityCalculation
    # from geoprob_pipe.calculations.system_calculations.piping_moria.dummy_input import PIPING_DUMMY_INPUT
    # import inspect
    #
    # reliability_calculation = PipingMORIASystemReliabilityCalculation(PIPING_DUMMY_INPUT)
    # sig = inspect.signature(reliability_calculation.given_system_variables_setup_function)
    # _ = [
    #     name for name, param in sig.parameters.items()
    #     if param.default == inspect.Parameter.empty and param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
    # ]

    # return ["mv_exit", "gamma_sat_deklaag"]  # TODO
    # return ["gamma_sat_deklaag"]  # TODO


def _expand(
        df_parameter_invoer_combined: DataFrame, df_identifiers: DataFrame, geopackage_filepath: str
) -> Dict[str, DataFrame]:

    # Add parameter invoer: op uittredepunten niveau
    required_input_parameters = _gather_required_input_parameters(geopackage_filepath=geopackage_filepath)
    collection_of_dfs: Dict[str, DataFrame] = {}
    for parameter_name in required_input_parameters:

        # Gather and merge input on uittredepunten-niveau
        df_gather = df_parameter_invoer_combined[
            (df_parameter_invoer_combined['parameter'] == parameter_name) &
            (df_parameter_invoer_combined['scope'] == 'uittredepunt')]
        df_gather = df_gather[["scope_referentie", "parameter_input"]]
        df = df_identifiers.copy(deep=True).merge(
            df_gather, how="left", left_on="uittredepunt_id", right_on="scope_referentie")
        df = df.drop(columns=["scope_referentie"])

        # Vak / scenario niveau
        df_gather = df_parameter_invoer_combined[
            (df_parameter_invoer_combined['parameter'] == parameter_name) &
            (df_parameter_invoer_combined['scope'] == 'vak') &
            (df_parameter_invoer_combined['ondergrondscenario_naam'].notna())]
        df_gather = df_gather[["scope_referentie", "ondergrondscenario_naam", "parameter_input"]]
        df_gather = df_gather.rename(columns={
            "scope_referentie": "vak_id",
            "ondergrondscenario_naam": "naam"})
        df['parameter_input'] = df['parameter_input'].combine_first(
            df_identifiers.copy(deep=True).merge(df_gather, on=["vak_id", "naam"], how="left")['parameter_input'])

        # Vak niveau
        df_gather = df_parameter_invoer_combined[
            (df_parameter_invoer_combined['parameter'] == parameter_name) &
            (df_parameter_invoer_combined['scope'] == 'vak') &
            (df_parameter_invoer_combined['ondergrondscenario_naam'].isna())]
        df_gather = df_gather[["scope_referentie", "parameter_input"]]
        df_gather = df_gather.rename(columns={"scope_referentie": "vak_id"})
        df['parameter_input'] = df['parameter_input'].combine_first(
            df_identifiers.copy(deep=True).merge(df_gather, on=["vak_id"], how="left")['parameter_input'])

        # Traject niveau
        df_gather = df_parameter_invoer_combined[
            (df_parameter_invoer_combined['parameter'] == parameter_name) &
            (df_parameter_invoer_combined['scope'] == 'traject')]
        if df_gather.__len__() >= 1:
            traject_value = df_gather['parameter_input'].values[0]
            df['parameter_input'] = df['parameter_input'].apply(lambda x: traject_value if isna(x) else x)

        # GIS spatial joins
        df_gather = df_parameter_invoer_combined[
            (df_parameter_invoer_combined['parameter'] == parameter_name) &
            (df_parameter_invoer_combined['scope'] == 'gis_uittredepunt')].copy(deep=True)
        df_gather = df_gather[["scope_referentie", "parameter_input"]]
        df_gather = df_gather.rename(columns={"scope_referentie": "uittredepunt_id"})
        df['parameter_input'] = df['parameter_input'].combine_first(
            df_identifiers.copy(deep=True).merge(df_gather, on=["uittredepunt_id"], how="left")['parameter_input'])

        # Add to collection
        collection_of_dfs[parameter_name] = df

    return collection_of_dfs


def _concat_collection(collection: Dict[str, DataFrame]):
    for parameter_name, df in collection.items():
        collection[parameter_name]['parameter_name'] = parameter_name
    return_df = concat([df for _, df in collection.items()], ignore_index=True)
    return_df = return_df.rename(columns={"naam": "ondergrondscenario_naam"})
    return return_df[["parameter_name", "vak_id", "uittredepunt_id", "ondergrondscenario_naam", "parameter_input"]]


def run_expand_input_tables(geopackage_filepath: str, add_frag_ref: bool = False) -> DataFrame:
    """ Performs logic to expand all parameter input from different levels to input on uittredepunt-level.

    :param geopackage_filepath:
    :param add_frag_ref: Indien True, dan wordt voor parameter input met een distribution_type 'cdf_curve' de referentie
        naar de invoer behouden. Deze is niet nodig voor de berekeningen, maar wordt wel gebruikt voor de visualisaties.
    :return: DataFrame with columns parameter_name, vak_id, uittredepunt_id, ondergrondscenario_naam and
        parameter_input.
    """

    # Construct df_parameter_invoer_combined
    tables = InputParameterTables(geopackage_filepath=geopackage_filepath)
    df_parameter_invoer_combined1 = _combine_parameter_invoer_sources(tables=tables)
    df_parameter_invoer_combined2 = _add_fragility_values_to_combined_parameter_invoer(
        df_parameter_invoer_combined=df_parameter_invoer_combined1, tables=tables,
        geopackage_filepath=geopackage_filepath, drop_ref=add_frag_ref==False)
    df_parameter_invoer_combined3 = _collect_right_columns_combined_parameter_invoer(
        df_parameter_invoer_combined=df_parameter_invoer_combined2)

    # Expand
    df_identifiers = _construct_df_identifiers(geopackage_filepath=geopackage_filepath, tables=tables)
    collection: Dict[str, DataFrame] = _expand(
        df_parameter_invoer_combined=df_parameter_invoer_combined3,
        df_identifiers=df_identifiers,
        geopackage_filepath=geopackage_filepath)

    return _concat_collection(collection=collection)



from typing import List
from pandas import DataFrame
import os
from dataclasses import dataclass
from geoprob_pipe.utils.validation_messages import BColors


@dataclass
class FailureQuery:
    query: str
    msg: str


class DataFrameQueryValidation:

    def __init__(self, df: DataFrame, failure_queries: List[FailureQuery]):
        self.df = df
        self.failure_queries = failure_queries

    def validate(self, export_dir: str, label_humanized: str) -> bool:
        for failure_query in self.failure_queries:

            # Assess
            df_failure_rows: DataFrame = self.df.query(failure_query.query)
            df_failure_rows = df_failure_rows.copy(deep=True)

            # Continue if no failures
            if df_failure_rows.__len__() == 0:
                continue

            # Prepare dataframe for export
            current_columns = df_failure_rows.columns
            new_column_order = ["validation_msg"]
            new_column_order.extend(current_columns)
            df_failure_rows['validation_msg'] = failure_query.msg
            df_failure_rows = df_failure_rows[new_column_order]

            # Report back
            os.makedirs(export_dir, exist_ok=True)
            export_path = os.path.join(export_dir, "df_failure_rows.xlsx")
            if os.path.exists(export_path):
                os.remove(export_path)
            df_failure_rows.to_excel(export_path)
            print(f"{BColors.WARNING}Validatie is (voortijdig) beëindigd omdat er {df_failure_rows.__len__()} "
                  f"validatie issues voor de '{label_humanized}'-tabel zijn gevonden. De gedetailleerde lijst is "
                  f"geëxporteerd naar onderstaande locatie. Los deze issues s.v.p. eerst op. \n"
                  f"{export_path}{BColors.ENDC}")
            return False

        return True



from pandas import DataFrame, concat
from typing import Optional, Dict, Union, List


class BColors:
    """ Helper object om prints kleur te geven. Gebruik als

    >>> print(BColors.WARNING, 'Mijn waarschuwing. ', BColors.ENDC)

    Let op. Met `BColors.ENDC` eindig je de kleur. Als je dit vergeet dan zijn alle volgende prints eveneens in kleur.

    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class ValidationMessages:

    def __init__(self, about: str = ""):
        self.about: str = about
        self.df: Optional[DataFrame] = None

    def _append_new_rows(self, new_rows: Dict):
        if self.df is None:
            self.df = DataFrame(new_rows)
        self.df = concat([self.df, DataFrame(new_rows)])

    @staticmethod
    def _to_list(msg: Union[str, List[str]]):
        assert isinstance(msg, (str, List))
        if isinstance(msg, str):
            msg = [msg]
        return msg

    @property
    def cnt(self) -> int:
        if self.df is None:
            return 0
        return self.df.__len__()

    def add_warning(self, msg: Union[str, List[str]]):
        msgs = self._to_list(msg=msg)
        new_rows = {
            "about": [self.about] * msgs.__len__(),
            "type": ["warning"] * msgs.__len__(),
            "msg": msgs,
        }
        self._append_new_rows(new_rows)

    def add_error(self, msg: Union[str, List[str]]):
        msgs = self._to_list(msg=msg)
        new_rows = {
            "about": [self.about] * msgs.__len__(),
            "type": ["error"] * msgs.__len__(),
            "msg": msgs,
        }
        self._append_new_rows(new_rows)

    def add_info(self, msg: Union[str, List[str]]):
        msgs = self._to_list(msg=msg)
        new_rows = {
            "about": [self.about] * msgs.__len__(),
            "type": ["info"] * msgs.__len__(),
            "msg": msgs,
        }
        self._append_new_rows(new_rows)

    def concat_with_df(self, df_to_append_to: Optional[DataFrame] = None) -> Optional[DataFrame]:
        """ Helper function to concatenate the validation messages with validation messages of another
        ValidationMessages-dataframe. """

        if self.df is None:
            return df_to_append_to

        if df_to_append_to is None:
            return self.df

        return concat([df_to_append_to, self.df])



from __future__ import annotations
from geoprob_pipe.cmd_app.general.project import created_project
from geoprob_pipe.cmd_app.general.geohydrologisch_model import created_model
from geoprob_pipe.cmd_app.spatial_layers import (
    added_dijktraject, added_vakindeling, added_uittredepunten, added_hrd, added_polderpeil, added_binnenteenlijn,
    added_buitenteenlijn, added_intredelijn, added_ahn)
from geoprob_pipe.cmd_app.spatial_joins import (
    coupled_hrd_to_uittredepunten, coupled_distances_to_uittredepunten, coupled_polderpeil_to_uittredepunten,
    coupled_uittredepunten_to_refline, coupled_uittredepunten_to_vakken, coupled_mv_exit_to_gis_parameter_invoer_table)
from geoprob_pipe.cmd_app.parameter_input.added_input_parameters import added_input_parameter_data
from typing import TYPE_CHECKING
from geoprob_pipe.cmd_app.run_calculations.run import run_calculations
import sys
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


EARLY_EXIT_MESSAGE = f"Applicatie vroegtijdig afgesloten"


def start_questionnaire(app_settings: ApplicationSettings):
    if not created_project(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    questionnaire(app_settings=app_settings)


def questionnaire(app_settings: ApplicationSettings):
    print(f"\nALGEMEEN")
    if not created_model(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)

    print(f"\nGIS LAGEN")
    if not added_dijktraject(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_vakindeling(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_hrd(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_uittredepunten(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_polderpeil(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_binnenteenlijn(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_buitenteenlijn(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not added_intredelijn(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    added_ahn(app_settings=app_settings, display_added_msg=True)  # AHN may be optional

    print(f"\nGEOGRAFISCHE KOPPELINGEN")
    if not coupled_uittredepunten_to_refline(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not coupled_hrd_to_uittredepunten(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not coupled_distances_to_uittredepunten(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not coupled_polderpeil_to_uittredepunten(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not coupled_uittredepunten_to_vakken(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)
    if not coupled_mv_exit_to_gis_parameter_invoer_table(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)

    print(f"\nPARAMETER INVOER")
    if not added_input_parameter_data(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)

    print(f"\nBEREKENINGEN UITVOEREN")
    if not run_calculations(app_settings=app_settings): sys.exit(EARLY_EXIT_MESSAGE)



from __future__ import annotations
import sqlite3
import json
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def _to_text(v):
    if v is None:
        return None
    if isinstance(v, (dict, list)):
        return json.dumps(v, ensure_ascii=False)
    if isinstance(v, (bool, int, float)):
        return json.dumps(v, ensure_ascii=False)
    return str(v)


def _upsert_metadata(conn, table, metadata_type, value):
    sql_update = f'UPDATE {table} SET "values" = ? WHERE metadata_type = ?'
    sql_insert = f'INSERT INTO {table} (metadata_type, "values") VALUES (?, ?)'

    with conn:  # transaction
        cur = conn.execute(sql_update, (value, metadata_type))
        if cur.rowcount == 0:
            conn.execute(sql_insert, (metadata_type, value))


def update_metadata(geoprob_pipe: GeoProbPipe):
    gpkg_path = geoprob_pipe.input_data.app_settings.geopackage_filepath
    table = "geoprob_pipe_metadata"

    # tekst setup voor variable values
    if geoprob_pipe.input_data.app_settings.to_run_vakken_ids:
        ran_vakken_ids = geoprob_pipe.input_data.app_settings.to_run_vakken_ids
    else:
        ran_vakken_ids = "all"

    n_calcs: int = len(geoprob_pipe.results.df_beta_scenarios_final)
    n_points: int = len(geoprob_pipe.results.df_beta_uittredepunten)
    n_vakken: int = len(geoprob_pipe.results.df_beta_vakken)
    ratio_points: float = n_points / len(geoprob_pipe.input_data.uittredepunten.gdf)

    records = [
        {"metadata_type": "last_calculation_run_datetime",
         "values": geoprob_pipe.input_data.app_settings.datetime_stamp},
        {"metadata_type": "last_calculation_rub_in_seconds",
         "values": geoprob_pipe.time_diff.total_seconds()},
        {"metadata_type": "last_calculation_run_vakken_to_run",
         "values": ran_vakken_ids},
        {"metadata_type": "last_calculation_run_nr_of_calculations",
            "values": n_calcs},
        {"metadata_type": "last_calculation_run_nr_of_uittredepunten",
            "values": n_points},
        {"metadata_type": "last_calculation_run_nr_of_vakken",
            "values": n_vakken},
        {"metadata_type": "last_calculation_run_percentage_of_uittredepunten",
            "values": ratio_points * 100}
    ]

    conn = sqlite3.connect(gpkg_path)

    for r in records:
        _upsert_metadata(conn=conn, table=table,
                         metadata_type=r["metadata_type"],
                         value=_to_text(r["values"]))
    conn.commit()
    conn.close()



from __future__ import annotations
from InquirerPy import inquirer
from typing import TYPE_CHECKING, Optional
import os
from geoprob_pipe.cmd_app.spatial_layers.uittredepunten.alg_walking_circles import algorithm_walking_circles
from shapely import Point
from geopandas import GeoDataFrame, read_file
import fiona
from geoprob_pipe.utils.validation_messages import BColors
import warnings
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def added_uittredepunten(app_settings: ApplicationSettings) -> bool:
    layers = fiona.listlayers(app_settings.geopackage_filepath)

    if "uittredepunten" not in layers:
        define_method_of_adding_uittredepunten(app_settings)
        return True

    gdf: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")
    if gdf.__len__() == 0:
        define_method_of_adding_uittredepunten(app_settings)
        return True

    print(BColors.OKBLUE, f"✔  Uittredepunten al toegevoegd ({gdf.__len__()} in totaal).", BColors.ENDC)
    add_maaiveld_niveau_exit_points()
    return True


def define_method_of_adding_uittredepunten(app_settings: ApplicationSettings):
    choices_list = ["Nu klikken in ArcGIS/QGIS", "Importeren uit GIS-bestand", "Automatische suggestie"]
    choice = inquirer.select(
        message="Er zijn nog geen uittredepunten toegevoegd. Hoe wil je deze toevoegen?",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        create_empty_uittredepunten_layer()
    elif choice == choices_list[1]:
        import_uittredepunten_gis_file(app_settings)
    elif choice == choices_list[2]:
        generate_uittredepunten_suggestions(app_settings=app_settings)
    else:
        raise ValueError


def add_maaiveld_niveau_exit_points():
    # sys.exit("Applicatie vroegtijdig afgesloten: Functie 'add_maaiveld_niveau_exit_points' nog niet afgerond.")
    # TODO
    return


def create_empty_uittredepunten_layer():
    raise NotImplementedError("Applicatie vroegtijdig afgesloten: Deze keuze is nog niet geïmplementeerd.")


def import_uittredepunten_gis_file(app_settings: ApplicationSettings):
    request_uittredepunten_filepath(app_settings=app_settings)


def import_from_geodatabase(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de layer waarin met de uittredepunten. Type 'listlayers' om "
                    "een overzicht te krijgen van de geodatabase-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geodatabase: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geodatabase: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf


def import_from_geopackage(filepath: str) -> GeoDataFrame:
    layer_name: Optional[str] = None
    layer_name_is_valid = False
    while layer_name_is_valid is False:
        layer_name: str = inquirer.text(
            message="Specificeer de layer waarin met de uittredepunten. Type 'listlayers' om "
                    "een overzicht te krijgen van de geopackage-layers. ",
        ).execute()

        layer_names = fiona.listlayers(filepath)
        layer_names.sort()
        layers_str = ", ".join(layer_names)
        if layer_name == "listlayers":
            print(BColors.OKBLUE, f"De volgende layers zijn beschikbaar in de geopackage: {layers_str}", BColors.ENDC)
            continue
        elif layer_name not in layer_names:
            print(BColors.OKBLUE, f"De laag name '{layer_name}' bestaat niet. De volgende layers zijn beschikbaar in "
                                  f"de geopackage: {layers_str}", BColors.ENDC)
            continue

        layer_name_is_valid = True

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Measured \\(M\\) geometry types are not supported.*")
        gdf: GeoDataFrame = read_file(filepath, layer=layer_name)
    return gdf


LEGAL_MODEL_OPTIONS =["Walking Circles", "Walking Circles, tilted"]


def generate_uittredepunten_suggestions(app_settings: ApplicationSettings):
    choices = inquirer.checkbox(
        message="Welk algoritme wil je gebruiken? Meerdere opties mogelijk.\n"
                "Press <space> to select, Enter when finished. ",
        choices=LEGAL_MODEL_OPTIONS,
    ).execute()

    if choices.__len__() == 0:
        print(BColors.OKBLUE, f"Je hebt geen selectie gemaakt. Maak een keuze of sluit af (ctrl+c).", BColors.ENDC)
        generate_uittredepunten_suggestions(app_settings=app_settings)

    if LEGAL_MODEL_OPTIONS[0] in choices:
        algorithm_walking_circles(app_settings=app_settings)
    if LEGAL_MODEL_OPTIONS[1] in choices:
        raise NotImplementedError(
            f"Applicatie vroegtijdig afgesloten: 'LEGAL_MODEL_OPTIONS[1]' is nog niet geïmplementeerd.")


def request_uittredepunten_filepath(app_settings: ApplicationSettings):

    # Request filepath
    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar de geopackage/shapefile/geodatabase waarin de "
                    "uittredepunten zitten.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not (filepath.endswith(".gpkg") or filepath.endswith(".shp") or filepath.endswith(".gdb")):
            print(BColors.WARNING, f"Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                                   f"eindigt op de extensie .{filepath.split(sep='.')[-1]}.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    # Import data
    if filepath.endswith(".shp"):
        gdf: GeoDataFrame = read_file(filepath)
    elif filepath.endswith(".gpkg"):
        gdf: GeoDataFrame = import_from_geopackage(filepath=filepath)
    elif filepath.endswith(".gdb"):
        gdf: GeoDataFrame = import_from_geodatabase(filepath=filepath)
    else:
        raise NotImplementedError(
            f"Applicatie vroegtijdig afgesloten: Een {filepath.split(sep='.')[-1]}-bestand is nog niet "
            f"geïmplementeerd.")

    # Confirm all are points
    all_geometries_are_points = gdf.geometry.apply(lambda geom: isinstance(geom, Point)).all()
    if not all_geometries_are_points:
        print(BColors.WARNING, f"Het geïmporteerde bestand bestaat niet (volledig) uit punten, maar ook uit "
                               f"andere typen geometrie. Enkel punten zijn toegestaan.", BColors.ENDC)
        request_uittredepunten_filepath(app_settings=app_settings)

    # Continue questionnaire
    specify_column_with_maaiveld_niveau(app_settings, gdf=gdf)


def specify_column_with_maaiveld_niveau(app_settings: ApplicationSettings, gdf: GeoDataFrame):
    column_name: Optional[str] = None
    column_name_is_valid = False
    while column_name_is_valid is False:
        column_name: str = inquirer.text(
            message="Specificeer de kolom waarin het maaiveld niveau staat. Type 'listcolumns' om "
                    "een overzicht te krijgen van de kolommen. Type 'n.a.' als het maaiveld niveau niet gekoppeld is "
                    "aan de punten, dan zal de applicatie in een volgende stap deze voor je downloaden.",
        ).execute()

        column_names = gdf.columns
        columns_str = ", ".join(column_names)
        if column_name == "listcolumns":
            print(BColors.OKBLUE,
                  f"De volgende kolommen zijn beschikbaar in de spatial layer: {columns_str}", BColors.ENDC)
            continue
        elif column_name == "n.a.":
            gdf_to_add = gdf[["geometry"]]
            gdf_to_add["mv_exit"] = -999.9
            gdf_to_add.to_file(app_settings.geopackage_filepath, layer="uittredepunten", driver="GPKG")
            print(BColors.OKBLUE, f"✅  Uittredepunten toegevoegd.", BColors.ENDC)
            return
        elif column_name not in column_names:
            print(BColors.OKBLUE, f"De kolom naam '{column_name}' bestaat niet. De volgende kolommen zijn beschikbaar "
                                  f"in de spatial layer: {columns_str}", BColors.ENDC)
            continue
        column_name_is_valid = True

    gdf_to_add = gdf[["geometry", column_name]]
    # No uittredepunt id yet. This will be set when the metering is determined.
    gdf_to_add = gdf_to_add.rename(columns={column_name: "mv_exit"})
    gdf_to_add.to_file(app_settings.geopackage_filepath, layer="uittredepunten", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Uittredepunten toegevoegd.", BColors.ENDC)



from typing import List
from geoprob_pipe.input_data.validation.dataframes.validation_objects import FailureQuery


FAILURE_QUERIES: List[FailureQuery] = [
    FailureQuery(
        query="distribution_type == 'cdf_curve' and fragility_values_ref.isnull()",
        msg="Geen referentie naar de fragility values gespecificeerd voor deze rij. Dit is vereist bij distributie "
            "type 'cdf_curve'."
    ),
    FailureQuery(
        query="distribution_type in ['normal', 'log_normal'] and mean.isnull()",
        msg="Geen gemiddelde waarde gespecificeerd voor deze rij. Dit is vereist bij distributie types 'normal' en "
            "'log_normal'."
    ),
    FailureQuery(
        query="distribution_type in ['normal', 'log_normal'] and variation.isnull() and deviation.isnull()",
        msg="Geen variatie coefficient of standaard deviatie gespecificeerd voor deze rij. Dit is vereist bij "
            "distributie types 'normal' en 'log_normal'."
    ),
    FailureQuery(
        query="parameter in ['d70', 'd70_m'] "
              "and distribution_type in ['deterministic', 'log_normal', 'normal'] "
              "and mean >= 0.001",
        msg="Parameter d70/d70_m moet een waarde hebben die kleiner is dan 0.001. Deze geef je op in de eenheid "
            "meters. Wellicht is de eenheid van de opgegeven waarde verkeerd?"
    ),
]



from __future__ import annotations
import os
import numpy as np
from pandas import DataFrame, concat, Series
from geopandas import GeoDataFrame
import plotly.colors as pc
from plotly.graph_objects import Figure, Scatter
from typing import TYPE_CHECKING, Dict, List, Tuple
from geoprob_pipe.cmd_app.parameter_input.expand_input_tables import run_expand_input_tables
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


TARGET_FREQS = np.array([
        0.1, 0.033333333, 0.01, 0.003333333, 0.001, 0.000333333, 0.0001, 3.33333E-05, 0.00001, 3.33333E-06])


def _collect_data(geoprob_pipe: GeoProbPipe) -> Tuple[DataFrame, GeoDataFrame, DataFrame, DataFrame]:

    # Collect uittredepunten
    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf

    # Collect physical values
    df = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(
        filter_deterministic=False, filter_derived=False)
    df = df[["uittredepunt_id", "ondergrondscenario_id", "vak_id", "variable", "distribution_type", "physical_value"]]
    df = df.merge(gdf_uittredepunten[["uittredepunt_id", "metrering", "hrd_name"]], on="uittredepunt_id", how="left")

    # Collect Beta results uittredepunten
    df_beta = geoprob_pipe.results.df_beta_uittredepunten

    # Collect hydra curves (from expansion to uittredepunt-level)
    df_input: DataFrame = run_expand_input_tables(
        geoprob_pipe.input_data.app_settings.geopackage_filepath, add_frag_ref=True)
    df_input = df_input[df_input["parameter_name"] == "buitenwaterstand"]
    df_input = concat(
        objs=[df_input.drop(columns=['parameter_input']),   # Removal of parameter_input column
              df_input['parameter_input'].apply(Series)], # Expansion of dict in parameter_input-column
        axis=1)

    return df, gdf_uittredepunten, df_beta, df_input


def _collect_hydra_curves(
        df: DataFrame, gdf_uittredepunten: GeoDataFrame) -> Dict[float, Dict[str, List[float]]]:
    hydra_curves = {freq: {"metrering": [], "level": []} for freq in TARGET_FREQS}

    for _, row in df.iterrows():
        if row["distribution_type"] != "deterministic":
            df_subset = gdf_uittredepunten[gdf_uittredepunten["vak_id"] == row["vak_id"]]
            if df_subset.empty:
                continue
            m_values = df_subset["metrering"].to_numpy()

            # Hydra exceedance curve
            freqs = np.array([fv.probability_of_failure for fv in row.loc["fragility_values"]], dtype=float)
            levels = np.array([fv.x for fv in row.loc["fragility_values"]], dtype=float)

            # Sort for interpolation
            sort_idx = np.argsort(freqs)
            freqs = freqs[sort_idx]
            levels = levels[sort_idx]

            # Interpolate levels for standard frequencies
            interp_levels = np.interp(TARGET_FREQS, freqs, levels)

            # Store values for each frequency
            for freq, level in zip(TARGET_FREQS, interp_levels):
                hydra_curves[freq]["metrering"].extend(m_values)
                hydra_curves[freq]["level"].extend(np.full_like(m_values, level))

    return hydra_curves


def _plot_continuous_exceedance_lines(
        fig: Figure, hydra_curves: Dict[float, Dict[str, List[float]]]):

    # Blue gradient for lines
    line_colors = pc.sample_colorscale(
        colorscale="Jet", samplepoints=np.linspace(start=0.2, stop=0.9, num=len(TARGET_FREQS)))
    freq_color_map = {f: c for f, c in zip(TARGET_FREQS, line_colors)}

    for freq, data in hydra_curves.items():
        # Sort by metrering for continuous line plotting
        sort_idx = np.argsort(data["metrering"])
        m_sorted = np.array(data["metrering"])[sort_idx]
        level_sorted = np.array(data["level"])[sort_idx]
        fig.add_trace(
            Scatter(
                x=m_sorted, y=level_sorted, mode="lines", line=dict(color=freq_color_map[freq], width=2),
                name=f"1/{1/freq:,.0f}".replace(",", "."),  # mark thousands
                showlegend=True))

    return fig


def _plot_physical_values(fig: Figure, df: DataFrame) -> Figure:
    beta_colorscale = "RdYlGn" # Green-to-red scale (low beta = red, high beta = green)
    betas = df["beta"].to_numpy()
    fig.add_trace(
        Scatter(
            x=df["metrering"], y=df["physical_value"],
            mode="markers", name="Physical values", showlegend=True,
            marker=dict(
                symbol="circle", size=7, color=betas, colorscale=beta_colorscale, cmin=1, cmax=10,
                line=dict(width=0.5, color="black"),
                colorbar=dict(
                    title="β", xanchor="right", ticks="outside",
                    x=-0.05,  # Make space on the left side for y-axis
                )),
            customdata=df[["beta"]],
            hovertemplate=("metering: %{x}<br>" +
                           "buitenwaterstand: %{y:.2f}<br>" +
                           "beta: %{customdata:.3f}")))

    return fig


def _update_layout(fig: Figure) -> Figure:
    fig.update_layout(
        title="Buitenwaterstanden bij herhaaltijd en physical values uit system design points",
        xaxis=dict(title="Metrering [m]", showgrid=True, gridwidth=0.5, gridcolor="gray"),
        yaxis=dict(title="Buitenwaterstand [m+NAP]", showgrid=True, gridwidth=0.5, gridcolor="gray"),
        legend=dict(
            orientation="v", yanchor="middle", y=0.5, xanchor="left", x=1.02, title="Legenda",
            bgcolor="rgba(255,255,255,0.7)"),
        legend_traceorder="reversed", height=600, margin=dict(r=200))
    return fig


def river_waterlevel(geoprob_pipe: GeoProbPipe, export: bool = False):

    # Prepare base data
    df, gdf_uittredepunten, df_beta, df_input = _collect_data(geoprob_pipe=geoprob_pipe)

    # Collect Hydra lines (grouped per frequency)
    hydra_curves: Dict[float, Dict[str, List[float]]] = _collect_hydra_curves(
        df=df_input, gdf_uittredepunten=gdf_uittredepunten)

    # Plot one continuous line per exceedance frequency
    fig = Figure()
    fig = _plot_continuous_exceedance_lines(fig=fig, hydra_curves=hydra_curves)

    # Buitenwaterstand markers with β color scale
    df_filtered = df[df["variable"] == "buitenwaterstand"]
    df_filtered = df_filtered.merge(df_beta[["uittredepunt_id", "beta"]], on="uittredepunt_id", how="left")
    fig = _plot_physical_values(fig=fig, df=df_filtered)

    # Layout
    fig = _update_layout(fig=fig)

    # Export
    if export:
        export_dir = os.path.join(geoprob_pipe.visualizations.graphs.export_dir, "grafiek_physical_values")
        os.makedirs(export_dir, exist_ok=True)
        fig.write_html(os.path.join(export_dir, "river_waterlevel.html"), include_plotlyjs='cdn')

    return fig


from __future__ import annotations
from geopandas import GeoDataFrame, read_file
from typing import TYPE_CHECKING
import sqlite3
from geoprob_pipe.cmd_app.utils.misc import get_geohydrological_model
from geoprob_pipe.calculations.systems.mappers.initial_input_mapper import INITIAL_INPUT_MAPPER
from pandas import DataFrame
import numpy as np
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def push_scenario_invoer_tabel(app_settings: ApplicationSettings):

    # Genereer tabel input (o.b.v. vakindeling)
    gdf_vakindeling: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="vakindeling")
    gdf_vakindeling = gdf_vakindeling.sort_values(by=["id"])
    df_scenarios = gdf_vakindeling[["id"]].copy()
    df_scenarios.loc[:, 'naam'] = "scenario1"
    df_scenarios.loc[:, 'kans'] = 1.00

    # Rename columns
    df_scenarios = df_scenarios.rename(columns={"id": "vak_id"})

    # Push to geopackage
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    df_scenarios.to_sql("scenario_invoer", conn, if_exists="replace", index=False)
    conn.close()


def push_parameter_invoer_tabel(app_settings: ApplicationSettings):

    model_string = get_geohydrological_model(app_settings=app_settings)

    # Start base of table
    df_dummy_data = DataFrame(INITIAL_INPUT_MAPPER[model_string]['input'])
    df_dummy_data = df_dummy_data.sort_values(by=["name"])
    df_parameter_invoer: DataFrame = df_dummy_data[[
        "name", "distribution_type", "mean", "variation", "deviation"]].copy()

    # Rename columns
    df_parameter_invoer = df_parameter_invoer.rename(columns={"name": "parameter"})

    # Convert distribution type objects to string
    df_parameter_invoer["distribution_type"] = df_parameter_invoer["distribution_type"].apply(lambda x: x.__str__())

    # Add missing columns
    df_parameter_invoer.loc[:, "scope"] = "traject"
    df_parameter_invoer.loc[:, "scope_referentie"] = ""
    df_parameter_invoer.loc[:, "ondergrondscenario_naam"] = ""
    df_parameter_invoer.loc[:, "bronnen"] = ""
    df_parameter_invoer.loc[:, "opmerking"] = ""
    df_parameter_invoer.loc[:, "fragility_values_ref"] = ""
    df_parameter_invoer.loc[:, "minimum"] = np.nan
    df_parameter_invoer.loc[:, "maximum"] = np.nan

    # Sort dataframe
    df_parameter_invoer = df_parameter_invoer[[
        "parameter", "scope", "scope_referentie", "ondergrondscenario_naam",
        "distribution_type", "mean", "variation", "deviation", "minimum", "maximum", "fragility_values_ref",
        "bronnen", "opmerking"]].copy()

    # Push to geopackage
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    df_parameter_invoer.to_sql("parameter_invoer", conn, if_exists="replace", index=False)
    conn.close()


def push_fragility_values_invoer_tabel(app_settings: ApplicationSettings):

    # Construct table
    water_levels = [3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8, 8.5, 9, 9.5, 10]
    exceedance_values = [5.8, 4.6, 2.8, 1.7, 1.3, 0.65, 0.377, 0.236, 0.119, 0.048, 0.0195, 0.0069, 0.00134, 0.0001]
    example_fragility_values = {
        "fragility_values_ref": ["voorbeeld_hr"] * water_levels.__len__(),
        "waarde": water_levels,
        "kans": exceedance_values}
    df_fragility_values_invoer = DataFrame(example_fragility_values)

    # Push to geopackage
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    df_fragility_values_invoer.to_sql("fragility_values_invoer", conn, if_exists="replace", index=False)
    conn.close()


DF_EMPTY_CORRELATIE_INVOER = DataFrame({"parameter_a": [], "parameter_b": [], "correlation": []})


def push_correlatie_invoer_tabel(app_settings: ApplicationSettings):
    # Push to geopackage
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    DF_EMPTY_CORRELATIE_INVOER.to_sql("correlatie_invoer", conn, if_exists="replace", index=False)
    conn.close()


def initiate_input_excel_tables(app_settings: ApplicationSettings):

    # Fill 'Scenario invoer'
    push_scenario_invoer_tabel(app_settings=app_settings)

    # Fill 'Parameter invoer'
    push_parameter_invoer_tabel(app_settings=app_settings)

    # Fill 'Fragility values invoer'
    push_fragility_values_invoer_tabel(app_settings=app_settings)

    # Fill 'Correlatie invoer'
    push_correlatie_invoer_tabel(app_settings=app_settings)



import sqlite3
from pandas import read_sql_query
from pandas import DataFrame
import numpy as np


def append_to_gis_join_parameter_invoer_table(df_sjoin: DataFrame, parameter_name: str, geopackage_filepath: str):
    """ Through mean only """

    assert df_sjoin.columns.__len__() == 2
    assert "mean" in df_sjoin.columns
    assert "uittredepunt_id" in df_sjoin.columns

    # Create other columns
    df_to_append = df_sjoin.copy(deep=True)
    df_to_append = df_to_append.rename(columns={"uittredepunt_id": "scope_referentie"})
    df_to_append["parameter"] = parameter_name
    df_to_append["scope"] = "uittredepunt"
    df_to_append["ondergrondscenario_naam"] = ""
    df_to_append["distribution_type"] = "deterministic"
    df_to_append["variation"] = np.nan
    df_to_append["deviation"] = np.nan
    df_to_append["minimum"] = np.nan
    df_to_append["maximum"] = np.nan
    df_to_append["fragility_values_ref"] = ""
    df_to_append["bronnen"] = ""
    df_to_append["opmerking"] = ""

    # Sort columns
    df_to_append = df_to_append[
        ["parameter", "scope", "scope_referentie", "ondergrondscenario_naam", "distribution_type", "mean", "variation",
         "deviation", "minimum", "maximum", "fragility_values_ref", "bronnen", "opmerking"]]

    # Get table names
    conn = sqlite3.connect(geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables_names = [row[0] for row in cursor.fetchall()]

    # If gis_join-table is non-existent yet
    if "gis_join_parameter_invoer" not in tables_names:
        df_to_append.to_sql("gis_join_parameter_invoer", conn, if_exists="replace", index=False)
        return
    # If polderpeil not yet in gis_join-table
    df_existing = read_sql_query("SELECT * FROM gis_join_parameter_invoer;", conn)
    if parameter_name not in df_existing['parameter'].unique():
        df_to_append.to_sql("gis_join_parameter_invoer", conn, if_exists="append", index=False)
        return
    raise ValueError
    # Should not have come here. Parameter should already be supplied, and thus user should not have been able to
    # make a new spatial join.


def append_hrd_to_gis_join_parameter_invoer_table(df_sjoin: DataFrame, geopackage_filepath: str):
    """ Through mean only """

    assert df_sjoin.columns.__len__() == 2
    assert "hrd_name" in df_sjoin.columns
    assert "uittredepunt_id" in df_sjoin.columns

    # Create other columns
    df_to_append = df_sjoin.copy(deep=True)
    df_to_append["parameter"] = "buitenwaterstand"
    df_to_append["scope"] = "uittredepunt"
    df_to_append = df_to_append.rename(columns={"uittredepunt_id": "scope_referentie"})
    df_to_append["ondergrondscenario_naam"] = ""
    df_to_append["distribution_type"] = "cdf_curve"
    df_to_append["mean"] = np.nan
    df_to_append["variation"] = np.nan
    df_to_append["deviation"] = np.nan
    df_to_append["minimum"] = np.nan
    df_to_append["maximum"] = np.nan
    df_to_append = df_to_append.rename(columns={"hrd_name": "fragility_values_ref"})
    df_to_append["bronnen"] = ""
    df_to_append["opmerking"] = ""

    # Sort columns
    df_to_append = df_to_append[
        ["parameter", "scope", "scope_referentie", "ondergrondscenario_naam", "distribution_type", "mean", "variation",
         "deviation", "minimum", "maximum", "fragility_values_ref", "bronnen", "opmerking"]]

    # Get table names
    conn = sqlite3.connect(geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables_names = [row[0] for row in cursor.fetchall()]

    # If gis_join-table is non-existent yet
    if "gis_join_parameter_invoer" not in tables_names:
        df_to_append.to_sql("gis_join_parameter_invoer", conn, if_exists="replace", index=False)
        return
    # If polderpeil not yet in gis_join-table
    df_existing = read_sql_query("SELECT * FROM gis_join_parameter_invoer;", conn)
    if "polderpeil" not in df_existing['parameter'].unique():
        df_to_append.to_sql("gis_join_parameter_invoer", conn, if_exists="append", index=False)
        return
    raise ValueError
    # Should not have come here. Parameter should already be supplied, and thus user should not have been able to
    # make a new spatial join.



r""".. _stationair-model:

Model 4a: Stationair model
==========================

Op deze pagina volgt een uitleg van de implementatie van het stationaire
model. Achtergronden staan in bijlage 4 van het Technisch rapport
Waterspanningen bij Dijken :cite:t:`trw_2004`.
Onder de aanname van horizontale stroming in het watervoerende zandpakket
en verticale stroming in de weerstand biedende deklaag is een analytische
oplossing beschikbaar voor het debiet en het stijghoogteverloop in het zand.

.. figure:: /_static/model4a_trwd.png
   :width: 100%

   Schematisering grondwaterstroming stationair model
   (figuur b4.4 uit :cite:t:`trw_2004`)


Het stationaire model houdt in dat de respons op een gegeven locatie
:math:`x` constant is. Op elke locatie is het verband tussen de stijghoogte
in het watervoerende pakket :math:`\phi(x)` en de respons :math:`r(x)`
bekend door:

.. math::

    \phi(x) = h_{ref} + r(x) (h_{rivier} - h_{ref})

en andersom:

.. math::

    r(x) = \frac{(\phi(x) - h_{ref})}{(h_{rivier} - h_{ref})}

Aan het zandpakket met afdekkende kleilaag worden de volgende weerstanden
gedefinieerd. Voor het voorland:

.. math::

    W_{1} = \lambda_{1} tanh(\frac{L_{1}}{\lambda_{1}})

en voor het achterland:

.. math::

    W_{3} = \lambda_{3} tanh(\frac{L_{3}}{\lambda_{3}})

De totale weerstand is de som van bovenstaande weerstanden:
:math:`\sum W = W_1 + L_2 + W_3`.

In het geval van een radiale (intrede)weerstand is er een extra
weerstandsfactor gedefinieerd:

.. math::

    W_{radiaal} = 0.44 D

De radiale weerstand kan aan de totale weerstand :math:`\sum W` worden
toegevoegd.

De respons ter plaatse van de buitenteen :math:`r_{but}` en de binnenteen
:math:`r_{bit}` van de dijk is gedefinieerd als:

.. math::

    r_{but} = \frac{(W_2 + L_2)}{\sum W}

en

.. math::

    r_{bit} = \frac{(W_3)}{\sum W}

Het model neemt aan dat tussen de buitenteen en de binnenteen van de dijk
er geen uitwisseling plaatsvindt tussen het zandpakket en de deklaag.
Over deze lengte :math:`L_2` wordt lineair geïnterpoleerd.
Een bijzonder geval is wanneer :math:`L_2 = 0`. We spreken dan van een
kantelpunt. De definitie van de respons in het kantelpunt (zie
bladzijde b3-7 van :cite:`trw_2004)` is:

.. math::

    r_{kp} = \frac{1}{(1 + \frac{W_1}{W_3})}

Voor de faalmechanismen piping en macrostabiliteit is het potentiaalverloop
verloop in de nabijheid van de dijk van belang. Door de aanname van lineaire
interpolatie wordt het potentiaalverloop beschreven door drie formules:
één voor het voorland tot de buitenteen, het gebied onder de dijk en een
formulering voor het achterland. Hiervoor is het nodig de :math:`x` positie
in het dwarsprofiel te kennen. Deze module hanteert hiervoor de
volgende definities:

De :math:`x` loopt op richting het achterland. De dijkzate is dan gedefinieerd
als: :math:`L_2 = x_{bit} - x_{but}`.

Als :math:`x < x_{but}` dan ligt :math:`x` in het voorland en als
:math:`x > x_{but}` dan ligt :math:`x` in het achterland.

Voorland:

.. math::

    r(x) = 1.0 - (1.0 - r_{but}) \frac{sinh(\frac{L_{1} + x - x_{but}}
    {\lambda_{1}})}{sinh (\frac{L_{1}}{\lambda_{1}})}

Onder de dijk:

.. math::

    r(x) = r_{bit} + (r_{but} - r_{bit}) \frac{x_{bit} - x}{L_{2}}

Achterland:

.. math::

    r(x) = r_{bit} \frac{sinh(\frac{L_{3} - x + x_{bit}}{\lambda_{3}})}
    {sinh (\frac{L_{3}}{\lambda_{3}})}

"""


import math
from dataclasses import dataclass
from typing import Tuple
from geoprob_pipe.calculations.physical_components.geohydro_functions import (
    calc_lambda, calc_r_BIT, calc_r_BUT, calc_W
    )


@dataclass
class Model4a:
    r"""Class for groundwater model 4A Technisch Rapport Waterspanningen
    bij Dijken.
    """

    kD: float
    D: float
    c1: float
    c3: float
    L1: float
    L3: float
    x_but: float
    x_bit: float

    # noinspection PyPep8Naming
    @property
    def L2(self) -> float:
        return abs(self.x_bit - self.x_but)

    @property
    def lambda1(self) -> float:
        return calc_lambda(self.kD, self.c1)

    @property
    def lambda3(self) -> float:
        return calc_lambda(self.kD, self.c3)

    # noinspection PyPep8Naming
    @property
    def W1(self) -> float:
        return calc_W(self.lambda1, self.L1)

    # noinspection PyPep8Naming
    @property
    def W3(self) -> float:
        return calc_W(self.lambda3, self.L3)

    # noinspection PyPep8Naming
    @property
    def W_rad(self) -> float:
        return 0.44 * self.D

    # noinspection PyPep8Naming
    @property
    def W_tot4a(self) -> float:
        return self.W1 + self.L2 + self.W3

    # noinspection PyPep8Naming
    @property
    def r_BUT(self) -> float:
        return calc_r_BUT(self.W1, self.L2, self.W3)

    # noinspection PyPep8Naming
    @property
    def r_BIT(self) -> float:
        return calc_r_BIT(self.W1, self.L2, self.W3)

    def respons(self, x: float) -> Tuple[float, float, float]:
        """calculate response at x given model
        x positive direction is inwards, so x_but < x_bit"""

        # Before floodplain
        if x < self.x_but - self.L1:
            r = 1.0

        # Floodplain
        elif self.x_but > x >= self.x_but - self.L1:
            r = 1.0 - (1.0 - self.r_BUT) * math.sinh(
                (self.L1 + x - self.x_but) / self.lambda1
            ) / math.sinh(self.L1 / self.lambda1)

        # Hinterland (where potential is affected)
        elif self.x_bit < x <= self.x_bit + self.L3:
            r = (
                self.r_BIT
                * math.sinh((self.L3 - x + self.x_bit) / self.lambda3)
                / math.sinh(self.L3 / self.lambda3)
            )

        #
        elif x > self.x_bit + self.L3:
            r = 0.0

        # dike
        else:
            r = self.r_BIT + (self.r_BUT - self.r_BIT) * (self.x_bit - x) / (
                self.x_bit - self.x_but
            )

        return r, self.r_BUT, self.r_BIT



from geoprob_pipe.calculations.systems.moria.initial_input import (
    INITIAL_INPUT as INITIAL_INPUT_MORIA)
from geoprob_pipe.calculations.systems.model4a.initial_input import (
    INITIAL_INPUT as INITIAL_INPUT_MODEL4A)
from geoprob_pipe.calculations.systems.wbi.initial_input import (
    INITIAL_INPUT as INITIAL_INPUT_WBI)

# TODO: Dynamisch maken? Forceren dat naamgeving overeenkomt en
# we dynamisch importeren?


INITIAL_INPUT_MAPPER = {
    # TODO: Dit is nog de oude model 4a logica. Aanpassen?
    "model4a": {
        "label": "Model 4a",
        "input": INITIAL_INPUT_MODEL4A,
    },
    "wbi": {
        "label": "WBI",
        "input": INITIAL_INPUT_WBI,
    },
    "moria": {
        "label": "MORIA",
        "input": INITIAL_INPUT_MORIA,
    },
}



from geoprob_pipe.app_object import GeoProbPipe
from geoprob_pipe.calculations.systems.base_objects.system_calculation import (
    SystemCalculation)
from geoprob_pipe.calculations.systems.single_calc import reproduce_single_calculation


from __future__ import annotations
from typing import TYPE_CHECKING
from plotly.graph_objects import Figure, Scatter
import os
from pandas import merge
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def physical_values_buitenwaterstand_and_top_zand(geoprob_pipe: GeoProbPipe, export: bool = False) -> Figure:

    # Get data for graphing
    df = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(
        filter_deterministic=False, filter_derived=True)
    df = df[["uittredepunt_id", "ondergrondscenario_id", "vak_id", "variable", "distribution_type", "physical_value"]]

    # Attach measure
    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf
    df = merge(
        left=df,
        right=gdf_uittredepunten[["uittredepunt_id", "metrering"]],
        on="uittredepunt_id",
        how="left")

    fig = Figure()

    df_filtered = df[df['variable'] == "buitenwaterstand"]
    fig.add_trace(
        Scatter(
            x=df_filtered['metrering'],
            y=df_filtered["physical_value"],
            name="Buitenwaterstand",
            mode='markers',
            marker=dict(symbol='circle', size=5, color='blue')))
    df_filtered = df[df['variable'] == "top_zand"]
    fig.add_trace(
        Scatter(
            x=df_filtered['metrering'],
            y=df_filtered["physical_value"],
            mode='markers',
            name="Top zand",
            marker=dict(symbol='circle', size=5, color='brown')))

    fig.update_layout(
        xaxis=dict(
            title=f"Metrering", type='linear', range=[df['metrering'].min()-10, df['metrering'].max()+10],
            showgrid=True, gridwidth=0.5, gridcolor="gray"),
        yaxis=dict(
            title=f"Buitenwaterstand en top zand [m+NAP]", showgrid=True, gridwidth=0.5, gridcolor="gray",
            minor=dict(showgrid=True, dtick=1))
    )

    if export:
        export_dir = geoprob_pipe.visualizations.graphs.export_dir
        os.makedirs(export_dir, exist_ok=True)
        fig.write_html(
            os.path.join(export_dir, f"physical_values_buitenwaterstand_and_top_zand.html"),
            include_plotlyjs='cdn')

    return fig
